.CT
Compiler Testing and Validation
.NH
Self Compilation
.LP
The main test of tcc was to see if it could compile itself.  This was
accomplished.  The self-compiled version of tcc was tested by using it
to compile itself again, producing a purely native version which the
system compiler had no part in producing.  The source files from this
first compilation were saved, and the compiler was recompiled yet
again.  The source files were compared, and found to be identical.
\fBAll\fP further testing of tcc was performed with a
second-or-greater generation self-compiled tcc, i.e. one which had
been compiled using the self-compiled version of itself.  Several small
debugging updates were also made by incrementally recompiling tcc with itself.
We believe that this is enough evidence to claim that tcc is self-hosting.
.LP
The Makefile which is supplied in the code listings with this report
is the one used to compile tcc with itself.
.NH
Supplied Test Cases
.LP
Each of the supplied test programs was compiled with tcc.  The
sources and assembly language outputs are given in the printout
section.  A typescript of the compilation and execution of the tests
is also provided.  Briefly, all tests compiled and executed correctly
except for the following:
.XP
externs.c \- did not work because floating point numbers are not
implemented in tcc.  A modified version, called ``externs-modified.c''
with the floating point types replaced by integral types, worked
correctly.
.XP
pointer.c \- compiled correctly but did not run, since absolute
addresses 0x20000 - 0x20100 are not readable by default on the VAX
Unix system.  A modified version, called ``pointer-modified.c'' was made
which reads 0x100 bytes starting at symbol ``main'' and this worked
correctly.
.XP
struct3.c \- used a structure before it was defined, and thus did not
compile.  The modified version (struct2.c) which does things in the
proper order worked correctly.
.KS
.NH
The CPR Utility
.LP
The cpr utility had one line in it which tcc could not handle.  It was
the following:
.DS
\f(CWenum langs {NONE, AUTO, C, FORTRAN, ICON, LISP} Language = AUTO;
.DE
The reason was that tcc has the documented limitation of not being
able to define an enumeration and use enumerators from it in
initializations in the same declaration.  When the line was changed to
read:
.DS
\f(CWenum langs {NONE, AUTO, C, FORTRAN, ICON, LISP} Language;
.DE
With an explicit initialization early in the mainline like this:
.DS
\f(CWLanguage = AUTO;
.DE
CPR was compiled with no further warnings or errors of any kind, and
the resulting version of CPR was used to print all listings
included with this report.
.KE
.NH
Obfuscated C Contest Entries
.LP
A number of winning entries from past years of the ``International
Obfuscated C Contest'' were run through tcc.  As expected, these
stressed the compiler more than most ``normal'' programs.  Some, which
primarily abused the C preprocessor, were not bothered with.  Others
caused huge cascades of error messages from tcc whose causes were not
determined.  One caused tcc to produce incorrect code which core
dumped when run.  Again, due to the nature of the programs, this was
not further investigated.  But eight programs did compile without
modification and ran correctly and printouts of these are included in
the test section of the listings.  There are two ``Hello World''
programs, one ASCII to Morse Code converter, one brutal abuse of the
``goto'' statement, a program which plays Othello, a program which
bubble-sorts the characters of the strings in its argument vector,
a program which scrolls a message across the display, and a program
which converts Arabic to Roman numerals.
.NH
Miscellaneous
.NH 2
Tetris
.LP
A particularly nice Obfuscated C Contest entry, which implements a
complete and playable game of Tetris, was made to work with tcc
courtesy of Robert Parker, a 2A Math student who has been instrumental
in the stress testing of tcc.  As seen in the comments in the source
code (included with the listings), only two minor changes had to be
made to get the program to compile, and one of those corrected a
violation of the ANSI standard.
.NH 2
Course Assignments
.LP
Two programs from computer engineering course assignments, totaling
perhaps 1500 lines of code, not counting comments, were compiled with
tcc.  As expected they compiled flawlessly.  The programs, which are a
cache simulator and a small database manager, are available online
with tcc and the rest of the tests, but have not been printed.
.NH 2
Bob Parker's Specials
.LP
Three test cases by Bob Parker (pr1.c, pr3.c, and swamp.c) may be
found in the listings.  These were verified to give the same results
when compiled with tcc as they do when compiled with the system compiler.
We think they stress the compiler a little more than the supplied test
cases do.
.KS
.NH 2
Public Domain Utilities
.LP
A few public domain Unix utilities were compiled with tcc and verified
to work, again by Bob Parker.  No modifications to the sources were
necessary.  At present, the list is the following:
.DS
\(bu stat (a directory lister)
\(bu tab  (a simple indent program)
\(bu weekday (a program to compute information about dates)
\(bu which (a program which shows what commands are aliased to)
.DE
The sources and executables as compiled with tcc are available
online.
.KE
.NH 2
Expression Stress Test
.LP
To test the the flushing of registers to the stack in the tcc code
generator, an extremely complex and ``deeply nested'' expression was
written.  It was verified that flushing registers to the stack was
required to keep all the temporaries around, and then in addition a number of
function calls were inserted into the expression to cause further
register flushes.  The expression was tested with a thousand inputs of
six random numbers each, generated with a program also compiled with
tcc.  The results were compared to those obtained when the expression
was compiled with the system compiler, and found to be identical.  This
test can be found in source files ``exprtest4.c'' and ``genrand.c''.
.KS
.CT
Compiler Performance
.LP
To evaluate the quality of code generation of the compiler, it was
compared to the system C compiler.  The system C compiler was
``handicapped'' by not allowing it to use register variables, reflecting
the fact that these were not implemented in tcc.  This was
done by using ``-Dregister='' as one of the compiler options.  Then the
entire source code for tcc was compiled with both the system compiler
and with tcc itself.  The results (stripped executables) were as
follows:
.DS
\f(CWtcc (self compiled):        95232 bytes
tcc (system compiled):      93184 bytes
.DE
.KE
.LP
It is seen that for this particular case, tcc output was 2.2% larger 
than system compiler output.  It is recognized, of course, that one can
get much better code by using register variables and the ``-O'' option with
the system compiler, and better code yet by using the GNU compiler,
but the goal here was to test the performance of a 3-month hack
compiler against the simplest ``real'' compiler available.
.KS
.LP
Next, the self compiled and system compiled versions of tcc were both
used to recompile one of the tcc source files (genexpr.c) and the runs
were timed.  The following results were obtained:
.DS
\f(CWtcc (self):        52.5u 3.2s 3:30 26% 126+369k 19+17io 7pf+0w
tcc (system):      51.3u 2.8s 1:41 53% 125+367k 28+19io 15pf+0w
.DE
.KE
.KS
.LP
While it is recognized that some of that time is spent in the C
preprocessor whose execution time is the same in both cases, it is not
unreasonable to assume that most of the time is spent in tcc itself,
and thus that the system compiler produces code which is only
marginally faster (approx. 3-5%) than that of tcc.  Again, the
performance increase of the system compiler code would likely be significant if
register variables and optimization were allowed.  Just for a test,
genexpr.c was again compiled, but with the system compiler itself (the
``-S'' option was used to prevent the assembly phase).
The times were as follows:
.DS
\f(CWsystem compiler:   35.1u 1.9s 1:01 59% 151+329k 31+28io 5pf+0w
.DE
It is seen that the system compiler compiles about 46% faster than tcc
does, though this would decrease if tcc were recompiled with register
variables (there are register declarations in the tcc source code).
.KE
