#include "exprstack.h"
#include "nodetypes.h"
#include "nodes.h"
#include "symtab.h"
#include "error.h"

static char *sccsid = "@(#) codegen.c 1.28@(#)";

extern int labelcnt;
extern long maxstack;
extern int labelwaiting;

int offsetlabel;

extern struct es_entry *es_push();
extern void es_pop();

extern int labelcnt;
extern long regmask;
extern struct es_entry expr_stack[];
extern int es_pointer;

/*
 *  Generate the prologue for a function.  Note: the list of registers to
 *  save is currently empty since the compiler doesn't do register variables.
 */
fn_prologue(name,globflag)
char *name;
int globflag;
{
	struct es_entry en1,en2;
	text();
	if(globflag) printf("\t.globl\t_%s\n",name);
	printf("\t.align\t1\n");
	printf("_%s:\n",name);
	printf("\t.word\t0\n");
	labelwaiting = 0;
}

/*
 *  Generate the instruction which decrements the stack pointer to make room
 *  for local variables.
 */
reservelocals()
{
	if(maxstack) {
		text();
		printf("\tsubl2\t$%ld,sp\n",maxstack);
		labelwaiting = 0;
	}
}

/*
 *  Generate the epilogue for a function.  Currently, this is just a "ret".
 */
fn_epilogue()
{
	text();
	printf("\tret\n");
	labelwaiting = 0;
}

/*
 *  Generate a label.  Make sure no two labels are on the same line.
 */
genlabel(i)
int i;
{
	if(labelwaiting++) putchar('\n');
	printf("L%d:",i);
}

/*
 *  This is the main entry point for the code generator.  It has three
 *  arguments:
 *
 *	 node - the tree node of the [compound] statement to compile
 *   lab1 - a label to jump to on a "break" instruction
 *   lab2 - a label to jump to on a "continue" instruction
 *
 *  lab1 and lab2 can be zero when a break/continue isn't valid.
 *
 *  The code generator is recursive and calls itself through this entry point
 *  to compile each individual statement.
 */
codegen(node,lab1,lab2)
struct treenode *node;
int lab1,lab2;
{
	struct es_entry en, *ep;
	struct treenode *n1;
	int l1,l2,l3,l4;
	int argcnt;

	if(!node) panic("codegen called with null pointer");

	switch(node->ntype) {
		case TN_WHILE:
			l1 = labelcnt++;
			l2 = labelcnt++;
			l3 = labelcnt++;
			genlabel(l1);
			gen_logicexpr(node->link2,l2,l3,l2);
			genlabel(l2);
			codegen(node->link3,l3,l1);
			en.type = ES_STATIC;
			en.u.val1 = l1;
			en.val2 = 0;
			geninst("jbr",1,&en);
			genlabel(l3);
			break;
		case TN_DOWHILE:
			l1 = labelcnt++;
			l2 = labelcnt++;
			l3 = labelcnt++;
			genlabel(l1);
			codegen(node->link2,l3,l2);
			genlabel(l2);
			gen_logicexpr(node->link3,l1,l3,l3);
			genlabel(l3);
			break;
		case TN_IF:
			l1 = labelcnt++;
			l2 = labelcnt++;
			gen_logicexpr(node->link2,l1,l2,l1);
			genlabel(l1);
			codegen(node->link3,lab1,lab2);
			genlabel(l2);
			break;
		case TN_IF_ELSE:
			l1 = labelcnt++;
			l2 = labelcnt++;
			l3 = labelcnt++;
			gen_logicexpr(node->link2,l1,l2,l1);
			genlabel(l1);
			codegen(node->link3->link1,lab1,lab2);
			en.type = ES_STATIC;
			en.u.val1 = l3;
			en.val2 = 0;
			geninst("jbr",1,&en);
			genlabel(l2);
			codegen(node->link3->link2,lab1,lab2);
			genlabel(l3);
			break;
		case TN_FOR:
			if(node->link2->link1) eval_and_toss(node->link2->link1);
			l1 = labelcnt++;
			l4 = labelcnt++;
			genlabel(l1);
			if(node->link2->link2) {
				l2 = labelcnt++;
				gen_logicexpr(node->link2->link2,l2,l4,l2);
				genlabel(l2);
			}
			if(node->link2->link3) l3 = labelcnt++; else l3 = l1;
			codegen(node->link3,l4,l3);
			if(node->link2->link3) {
				genlabel(l3);
				eval_and_toss(node->link2->link3,0);
			}
			en.type = ES_STATIC;
			en.u.val1 = l1;
			en.val2 = 0;
			geninst("jbr",1,&en);
			genlabel(l4);
			break;		
		case TN_BREAK:
			if(lab1) {
				en.type = ES_STATIC;
				en.u.val1 = lab1;
				en.val2 = 0;
				geninst("jbr",1,&en);
			} else errl(NOBREAK,node);
			break;
		case TN_CONTINUE:
			if(lab1) {
				en.type = ES_STATIC;
				en.u.val1 = lab2;
				en.val2 = 0;
				geninst("jbr",1,&en);
			} else errl(NOCONT,node);
			break;
		case TN_RETURN:
			if(node->link2) {
				en.type = ES_REG;
				en.u.val1 = 0;
				if(!genexpr(node->link2,&en,4)) {
					ep = &expr_stack[es_pointer];
					if(ep->type != ES_REG || ep->u.val1 != 0) 
						geninst("movl",2,ep,&en);
					es_pop(1);
				}
			}
			printf("\tret\n");
			labelwaiting = 0;
			break;
		case TN_SWITCH:
			handle_switch(node,lab2);
			break;
		case TN_EXPR_STAT:
			if(node->link2) eval_and_toss(node->link2);
			break;
		case TN_GOTO:
			n1 = node->link2->v_defnode;
			if(n1->v_labelno) {
				en.type = ES_STATIC;
				en.u.val1 = n1->v_labelno;
				en.val2 = 0;
				geninst("jbr",1,&en);
			} else errl(GOTOUNDEFDLABEL,node);
			break;
		case TN_LABEL:
			n1 = node->link2;
			switch(n1->ntype) {
				case TN_CASE:
					if(!n1->link2->v_uval) errl(NOTINSWITCH,n1,"case");
					else genlabel(n1->link2->v_uval);
					break;
				case TN_DEFAULT:
					if(!n1->link1->v_uval) errl(NOTINSWITCH,n1,"default");
					else genlabel(n1->link1->v_uval);
					break;
				case TN_GOTOLABEL:
					genlabel(n1->v_defnode->v_labelno);
					break;
				default:
					panic("bad label node");
			}
			codegen(node->link3,lab1,lab2);
			break;
		case TN_COMPOUNDSTAT:
			n1 = node->link3;
			while(n1) {
				codegen(n1,0,0);
				n1 = n1->link1;
			}
			n1 = node->link2;
			while(n1) {
				codegen(n1,lab1,lab2);
				n1 = n1->link1;
			}
			break;
		default:
			panic("unknown tree node on codegen");
	}
}

/*
 *  The compiler dumps the register contents on the stack on function calls
 *  or when it runs out of registers.  Wherever possible, it incorporates
 *  the "stack pop" necessary later into the arguments of an instruction.
 *  This routine unstacks things which would not be popped by a longword
 *  instruction, to prevent the stack pointer from getting messed up.
 */
stackfix(what,size)
struct es_entry *what;
int size;
{
	struct es_entry en;
	int reg;
	if(what->type == ES_POSTINCR && what->u.val1 == REG_SP && size!=4) {
		reg = getreg();
		en.u.val1 = 1<<reg;
		en.type = ES_CONST;
		geninst("popr",1,&en);
		what->type = ES_REG;
		what->u.val1 = reg;
	}
}
