#include "def.h"
#include "nodes.h"
#include "nodetypes.h"
#include "types.h"
#include "error.h"
#include "symtab.h"
static char *sccsid = "@(#) decl.c 1.40@(#)";

#define TTS_CHAR 		1
#define TTS_SHORT		2
#define TTS_INT			4
#define TTS_LONG		8
#define TTS_UNSIGNED	16
#define TTS_SIGNED		32
#define TTS_VOID		64

/*
 *  This function builds the type expression for a structure.  It modifies
 *  an already allocated type node.
 *
 *  tnode - pointer to a sruct type to turn into TT_STRUCTHEADER
 *
 *  node - pointer to a TN_STRUCT, in which
 *         link1->pointer to TN_STRUCT or TN_UNION
 *         link2->pointer to structure name node
 *         link3->pointer to list of TN_SDECLARATION
 *
 *  Some gymnastics are done to ensure that structure files are always aligned
 *  to a multiple of their size, and that this remains true when the structure
 *  is placed in an array.
 */

void build_struct(tnode,node)
struct type *tnode;
struct treenode *node;
{
	register struct type *t1, *t2, *t3;
    register struct member *m1, *m2 = 0;
	register struct treenode *p1 = node->link3;
	register struct treenode *p2;
	struct treenode *p3;
	int uflag = (node->link1->ntype == TN_UNION);
	int sc,ar;
	long offset = 0;
	long siz;

	tnode->type = uflag ? TT_UNIONHEADER : TT_STRUCTHEADER;

	tnode->tu.su.alignreq = tnode->tu.su.size = 0;
	/*
     *  Process the chain of TN_SDECLARATION hanging off node->link3;
	 */
    while(p1) {
		t2 = process_declaration_specs(p1->link2,&sc,0);
		/*
		 *  Now process all the declarators hanging off p1->link3;
		 */
		p2 = p1->link3;
		while(p2) {
			if(p2->link3) {
				errl(NOBITFIELDS,p2->link3);
			} else {
				p3 = p2->link2;
				t1 = buildtypechain(&p3,t2,0);
				if(!p3) panic("symbol missing in declarator");
				/*
				 *  t1 -> type chain for current structure member
				 *  p3 -> pointer to TN_ID
				 */
				t3 = skiprefnodes(t1);
				if(t3->type == TT_FUNCTIONRETURNING) {
					errl(FUNCMEMBER,p3,p3->v_str);
					freetype(t1);
				} else {
					m1 = (struct member *) safealloc(sizeof(struct member));

					if(!m2) tnode->tn.members = m1; else m2->next = m1;
					m2 = m1;
				
					m1->name = save_str(p3->v_str);
					m1->type = t1;
					siz = size(t1);
					if(!siz) {
						errl(DECLSIZE,p2);
						m1->offset = 0;
					} else {
						ar = alignreq(t1);
						if(ar > tnode->tu.su.alignreq)
							tnode->tu.su.alignreq = ar;

						if(uflag) {
							m1->offset = 0;
							if(siz > offset) offset = siz;
						} else {
							offset = (offset+(1<<ar)-1) & (~0)<<ar;
							m1->offset = offset;
							offset += siz;
						}
					}
				}
			}
			p2 = p2->link1;
		}
		freetype(t2);
		p1 = p1->link1;
    }
	if(m2) m2->next = 0; else tnode->tn.members = 0;

	if(tnode->tn.members) tnode->tu.su.size =
		(offset+(1<<tnode->tu.su.alignreq)-1) & (~0)<<tnode->tu.su.alignreq;
}

/*
 *  This function handles a structure declarator, i.e. a declaration specifier
 *  of the form "struct ...".  It manages the structure definitions in the
 *  symbol table.  The arguments are:
 *
 *   node:    Tree node of type TN_STRUCT, with children as follows:
 *             link1 -> TN_STRUCT or TN_UNION node to indicate whether a
 *			  		    "struct" or "union" keyword was given
 *			   link2 -> (optional) string node with the name, if given
 *			   link3 -> (optional) list of structure field declarations.
 *
 *   tnode:   The type node to turn into a TT_STRUCTHEADER node.
 * 
 *   rhsflag: A kludge, indicating that the structure should be put in the
 *			  symbol table even when it is incompletely defined and already
 *			  exists there.  This is to implement an irregularity in the
 *			  C language, allowing structure definitions to be hidden inside
 *			  inner scopes by simply supplying an incomplete definition.
 */

void handle_struct(node,tnode,rhsflag)
int rhsflag;
struct treenode *node;
struct type *tnode;
{
	struct symbol *sym;

	if(node->link2) {
		sym = find_sym(node->link2->v_str,ET_STRUCTDEF,0);

		if(sym) {
			/*
			 *  Structure is in the symbol table.
			 */
			if(node->link3) {
				/*
				 *  This instance has a member list.
				 */
				if(sym->type->tn.members) {
					/*
					 *  But so does the structure in the symbol table.
					 */
					if(blocklevel == sym->blocklevel) {
						/*
						 *  Prevent attempts to redefine a structure at the
						 *  same block level.
						 */
						errl(STRUCTREDEF,node->link3,node->link2->v_str);
						tnode->type = TT_TYPEREF;
						tnode->tn.next = sym->type;
						sym->type->refcount++;
					} else {
						/*
						 *  A structure definition from an outer scope is
						 *  being hidden.
						 */
						sym = make_sym(blocklevel,
							node->link2->v_str,ET_STRUCTDEF);
						sym->type = tnode;
						tnode->refcount++;
						build_struct(tnode,node);
					}
				} else {
					/*
					 *  The incomplete structure in the symbol table is being
					 *  completed.
					 */
					build_struct(sym->type,node);
					tnode->type = TT_TYPEREF;
					tnode->tn.next = sym->type;
					sym->type->refcount++;
				}
			} else {
				/*
				 *  Structure is in the symbol table, and this declaration
				 *  has no member list.
				 */
				if(sym->blocklevel < blocklevel && rhsflag) {
					sym = make_sym(blocklevel,node->link2->v_str,ET_STRUCTDEF);
					sym->type = tnode;
					tnode->refcount++;
					build_struct(tnode,node);
				} else {
					tnode->type = TT_TYPEREF;
					tnode->tn.next = sym->type;
					sym->type->refcount++;
				}
			}
		} else {
			/*
			 *  Structure does not exist in symbol table.  Add it, with
			 *  or without member nodes at present.
			 */
			sym = make_sym(blocklevel,node->link2->v_str,ET_STRUCTDEF);
			sym->type = tnode;
			tnode->refcount++;
			build_struct(tnode,node);
		}
	} else {
		/*
		 *  No identifier given, therefore there must be a structure member
		 *  list (this is enforced by the grammar).
		 */
		build_struct(tnode,node);
	}
}

/*
 *  This function is very analogous to build_struct, refer to comments there.
 *
 *  node -> list of TN_ENUM, each of which has
 *           link2 -> identifier node
 *           link3 -> constant expression or null
 */

void build_enum(tnode,node)
struct type *tnode;
struct treenode *node;
{
	extern long get_const();
	long count = 0;
	struct enumerator *e1, *e2 = 0;

	tnode->type = TT_ENUMHEADER;
	tnode->tn.enumlist = 0;

    while(node) {
		e1 = (struct enumerator *) safealloc(sizeof(struct enumerator));
		e1->name = save_str(node->link2->v_str);
		if(!e2) tnode->tn.enumlist = e1; else e2->next = e1;
		e2 = e1;
		if(node->link3) count = get_const(node->link3);
 		e1->value = count++;
		node = node->link1;
	}
}

/*
 *  This function is analogous to handle_struct, see comments earlier.  Simpler
 *  rules apply though, so it is shorter.
 *
 *  node -> TN_ENUM, which has
 *          link1 -> identifier node or null
 *          link2 -> linked list of enumerators or null
 */

void handle_enum(node,tnode)
struct treenode *node;
struct type *tnode;
{
	struct symbol *sym;
	if(node->link1) {
		sym = find_sym(node->link1->v_str,ET_ENUMDEF,0);
		if(!sym && !node->link2) {
			errl(ENUMNOTDEF,node->link1,node->link1->v_str);
			tnode->type = TT_ENUMHEADER;
			tnode->tn.enumlist = 0;
		} else if(sym) {
			if(node->link2) {
				if(blocklevel > sym->blocklevel) {
					build_enum(tnode,node->link2);
					sym = make_sym(blocklevel,node->link1->v_str,ET_ENUMDEF);
					sym->type = tnode;
					tnode->refcount++;
				} else {
					errl(ENUMREDEF,node->link2,node->link1->v_str);
					tnode->type = TT_TYPEREF;
					tnode->tn.next = sym->type;
					sym->type->refcount++;
				}
			} else {
				tnode->type = TT_TYPEREF;
				tnode->tn.next = sym->type;
				sym->type->refcount++;
			}
		} else {
			build_enum(tnode,node->link2);
			sym = make_sym(blocklevel,node->link1->v_str,ET_ENUMDEF);
			sym->type = tnode;
			tnode->refcount++;
		}
	} else {
		build_enum(tnode,node->link2);
		sym = make_sym(blocklevel,"",ET_ENUMDEF);
		sym->type = tnode;
		tnode->refcount++;
	}
}

/*
 *  This routine processes a string of declaration specifiers, such as
 *
 *  "static const unsigned short int"
 *
 *  It makes sense of this mess, ensures that only valid combinations of
 *  keywords are given, and returns a type node corresponding to what the
 *  declaration specifiers mean.  Arguments:
 *
 *   ptr:      Pointer to linked list of tree nodes representing the keywords
 *   sc_ptr:   Pointer to an integer in which to store a storage class, if
 *			   one is found.
 *   rhsflag:  Kludge flag to be passed to handle_struct, if a "struct"
 *             keyword is in the declaration specifiers.
 */
struct type *process_declaration_specs(ptr,sc_ptr,rhsflag)
int rhsflag;
register struct treenode *ptr;
int *sc_ptr;
{
	struct symbol *sym;
	int scflag = 0;
	int tsmask = 0;
	struct treenode *savedptr;
	register struct type *tnode = (struct type *)safealloc(sizeof(struct type));
	tnode->refcount = 1;

	tnode->type = TT_BASICTYPE;
	tnode->tn.next = 0;
	tnode->tu.flags = 0;

	while(ptr) {
		switch(ptr->ntype) {
			case TN_STORAGECLASS:
				if(scflag) errl(ERR_SYNTAX,ptr); else {
					*sc_ptr = ptr->link1->ntype;
					scflag = 1;
				}
				break;
			case TN_TYPESPECIFIER:
				if(tnode->type != TT_BASICTYPE) errl(ERR_SYNTAX,ptr->link1);
				else switch(ptr->link1->ntype) {
					case TN_CHAR:
						if(tsmask &
							TTS_CHAR+TTS_SHORT+TTS_INT+TTS_LONG+TTS_VOID)
						errl(ERR_SYNTAX,ptr->link1); else tsmask |= TTS_CHAR;
						break;
					case TN_SHORT:
						if(tsmask & TTS_CHAR+TTS_LONG+TTS_SHORT+TTS_VOID)
						errl(ERR_SYNTAX,ptr->link1); else tsmask |= TTS_SHORT;
						break;
					case TN_INT:
						if(tsmask & TTS_CHAR+TTS_INT+TTS_VOID)
						errl(ERR_SYNTAX,ptr->link1); else tsmask |= TTS_INT;
						break;
					case TN_LONG:
						if(tsmask & TTS_CHAR+TTS_LONG+TTS_SHORT+TTS_VOID)
						errl(ERR_SYNTAX,ptr->link1); else tsmask |= TTS_LONG;
						break;
					case TN_UNSIGNED:
						if(tsmask & TTS_UNSIGNED+TTS_SIGNED+TTS_VOID)
						errl(ERR_SYNTAX,ptr->link1);else tsmask |= TTS_UNSIGNED;
						break;
					case TN_SIGNED:
						if(tsmask & TTS_UNSIGNED+TTS_SIGNED+TTS_VOID)
						errl(ERR_SYNTAX,ptr->link1); else tsmask |= TTS_SIGNED;
						break;
					case TN_VOID:
						if(tsmask & TTS_CHAR+TTS_SHORT+TTS_INT+TTS_LONG+
						TTS_UNSIGNED+TTS_SIGNED) errl(ERR_SYNTAX,ptr->link1);
						else tsmask |= TTS_VOID;
						break;
					case TN_STRUCT:
						if(tsmask) errl(ERR_SYNTAX,ptr->link1); else
							handle_struct(ptr->link1,tnode,rhsflag);
						break;
					case TN_TYPEID:
						if(tsmask) errl(ERR_SYNTAX,ptr->link1); else {
							sym = find_sym(ptr->link1->v_str,ET_TYPEDEF,0);
							if(!sym) panic("typedef vanished");
							tnode->type = TT_TYPEREF;
							sym->type->refcount++;
							tnode->tn.next = sym->type;
						}
						break;
					case TN_ENUM:
						if(tsmask) errl(ERR_SYNTAX,ptr->link1); else
							handle_enum(ptr->link1,tnode);
						break;
					/* structure/union stuff still missing */
					default:
						panic("bad type spec");
				}
				break;
			case TN_TYPEQUALIFIER:
				{
					int tq_mask;
					switch(ptr->link1->ntype) {
						case TN_CONST:
							tq_mask = QF_CONST;
							break;
						case TN_VOLATILE:
							tq_mask = QF_VOLATILE;
							break;
						default:
							panic("bad type qual");
					}
					if(tnode->tu.flags & tq_mask) errl(ERR_SYNTAX,ptr->link1);
					tnode->tu.flags |= tq_mask;
				}
				break;
			default:
				panic("bad decl spec");
		}
		savedptr = ptr;
		ptr = ptr->link2;
	}
	/*
	 *  So far, we have a type node which is one of the following:
	 *    TT_BASICTYPE,TT_TYPEREF,TT_STRUCHEADER,TT_ENUMHEADER, TT_UNIONHEADER
	 *
	 *  Type qualifier flags are in tnode->flags, and type specifier flags
	 *  are in tsmask.
	 */
    if(tnode->type == TT_BASICTYPE) {
		if(tsmask & TTS_UNSIGNED) tnode->tu.flags |= TF_UNSIGNED;
		if(tsmask & TTS_CHAR) tnode->tu.flags |= TF_BIT8;
		else if(tsmask & TTS_SHORT) tnode->tu.flags |= TF_BIT16;
		else if(tsmask & TTS_LONG) tnode->tu.flags |= TF_BIT32;
		else if(tsmask & TTS_UNSIGNED+TTS_SIGNED+TTS_INT)
			tnode->tu.flags |= TF_BIT32;
		else if(tsmask & TTS_VOID) tnode->tu.flags |= TF_VOID;
		else tnode->tu.flags |= TF_BIT32;
	} else if(tsmask) errl(ERR_SYNTAX,savedptr);

	return tnode;
}

/*
 *  This function prefixes a type expression with another type constructer
 *  such as TT_POINTERTO, TT_ARRAYOF, ...  It sets the reference count of
 *  the new head node to 1, but does not increment the reference count of
 *  the old head node.
 */
void maketypenode(head,type)
struct type **head;
int type;
{
	struct type *new = (struct type *) safealloc(sizeof(struct type));
	new->type = type;
	new->refcount = 1;
	new->tn.next = *head;
	*head = new;
}

/*
 *  This routine is a toy version of process_declaration_specs which allows
 *  just const/volatile qualifiers, used to qualify a pointer.
 */
int typequalifiers(node)
struct treenode *node;
{
	int tb,tq = 0;
	while(node) {
		switch(node->link2->ntype) {
			case TN_CONST:
				tb = QF_CONST;
				break;
			case TN_VOLATILE:
				tb = QF_VOLATILE;
				break;
			default:
				panic("bad type qual");	
		}
		if(tb & tq) errl(ERR_SYNTAX,node); else tq |= tb;
		node = node->link1;
	}
	return tq;
}

/*
 *  This function skips the "type ref" nodes inserted into type chains when
 *  they join another type chain.  These were originally intended to aid in
 *  deallocating the chains, but this is now done with reference counts.  The
 *  sole reason for keeping them (and putting a billion skiprefnodes() calls
 *  into the code) was to have a place to store const/volatile qualifiers
 *  when several different objects share the same type, such as here:
 *
 *  struct {int a,b } * const c, * volatile d;
 *
 *  Too bad that we never had the time/energy to actually implement
 *  const/volatile checking in the semantic analyzer.
 */
struct type *skiprefnodes(ptr)
struct type *ptr;
{
	if( ptr == NULL ) return NULL;
	while(ptr->type == TT_TYPEREF) ptr = ptr->tn.next;
	return ptr;
}

/*
 *  Another central function.  This one processes the junk surrounding a
 *  declarator, that is, all the "()", "*", and "[]" stuff in the example below:
 *
 *  int (*a[])();
 *
 *  It takes a head node, given by t1, and returns the head node of the
 *  extended type chain.  p1 and trailptr point to pointers to parse tree
 *  nodes; these are passed by reference so that the caller can retrieve
 *  the updated versions.  On exit, p1 points to the identifier node at the
 *  core of the declarator, or is null if the declarator is missing.
 *  trailptr lags one node behind p1.  It can be given as zero in which case
 *  it is not updated.
 */
struct type *buildtypechain(p1,t1,trailptr)
register struct treenode **p1, **trailptr;
struct type *t1;
{
	int ar;
	long siz;
	register struct type *t2;
	t1->refcount++;
	while(*p1) {
		if((*p1)->ntype == TN_ID) break;
		switch((*p1)-> ntype) {
			case TN_SUBSCRIPT:
				t2 = skiprefnodes(t1);
				if(t2->type == TT_FUNCTIONRETURNING)
					errl(ARRAYOFFUNCTION,*p1);
				else if(t2->type == TT_BASICTYPE && t2->tu.flags & TF_VOID)
					errl(ARRAYOFVOID,*p1);
				else {
					if(t2->type == TT_ARRAYOF && !t2->tu.array.numentries)
						errl(ARRAYDIMS,*p1);
					maketypenode(&t1,TT_ARRAYOF);
					if((*p1)->link2) {
						t1->tu.array.numentries = get_const((*p1)->link2);
					} else t1->tu.array.numentries = 0;

					siz = size(t2);

					if(!siz) {
						errl(ARRAYOFINCOMP,*p1);
						t1->tu.array.sizeeach = 1;
					} else {
						t1->tu.array.sizeeach = siz;
					}
				}
				break;
			case TN_FUNCTIONDECL:
			case TN_FNDECL_TYPELIST:
			case TN_FNDECL_IDLIST:
				t2 = skiprefnodes(t1);
				if(t2->type == TT_FUNCTIONRETURNING)
					errl(FUNCTIONRETFUNCTION,*p1);
				else if(t2->type == TT_ARRAYOF)
					errl(FUNCTIONRETARRAY,*p1);
				else
					maketypenode(&t1,TT_FUNCTIONRETURNING);
				break;
			case TN_POINTER:
				maketypenode(&t1,TT_POINTERTO);
				t1->tu.flags = 0;
				break;
			case TN_QUALPOINTER:
				maketypenode(&t1,TT_POINTERTO);
				t1->tu.flags = typequalifiers((*p1)->link2);
				break;
			default:
				panic("bad declarator");
		}
		if(trailptr) *trailptr = *p1;
		*p1 = (*p1)->link1;
	}
	return t1;
}

/*
 *  This function processes a declaration node, be this a function argument
 *  declaration, a global variable declaration, or a local variable declaration.
 */
process_declaration(node)
struct treenode *node;
{
	int sc, sc1 = -1;
	struct treenode *p1, *p2;
	register struct treenode *ptr;
	long siz;
	int ar;
	struct type *t1;
	register struct type *t2;
	register struct symbol *sym;
	int label;

    struct type *tnode=process_declaration_specs(node->link2,&sc1,!node->link3);

	ptr = node->link3;

	while(ptr) {
		sc = sc1;
		p1 = ptr->link2;
		t1 = buildtypechain(&p1,tnode,0);
		t2 = skiprefnodes(t1);
		if(!p1) panic("symbol missing in declarator");
		/*
		 *  t1			- pointer to type chain
		 *  t2			- as above with ref nodes skipped
		 *  sc			- storage class (-1 if not given)
		 *  ptr->link3  - pointer to initializer node, if any
		 *  p1->v_str		- name of the identifier being declared
		 *
		 *  First of all, if we are at block level > 0, and the identifier
		 *  is already defined for this block level, then we have an error.
		 *  Structure/union and enumeration definitions are exempt from this.
		 */
		sym = find_sym(p1->v_str,~(ET_STRUCTDEF+ET_ENUMDEF),blocklevel);

		if(blocklevel && sym) {
			errl(REDECLARATION,p1,p1->v_str);
			freetype(t1);
		}
		/*
		 *  If we're still here, then one of the following is true:
		 *   (a) we are global (blocklevel = 0) and sym points to any previous
		 *       global definition of this identifier.
		 *   (b) we are local, and the identifier has not been declared at
		 *       this block level at all.
		 */
		else if(t2->type == TT_FUNCTIONRETURNING && sc != TN_TYPEDEF) {
			if(sc<0 || sc == TN_EXTERN || sc == TN_STATIC) {
				if(ptr->link3) errl(ERR_SYNTAX,ptr->link3);
				if(sc<0) sc = TN_EXTERN;
				/*
			     *  We have a function predeclaration.  The storage class is
				 *  register or extern.  Check for conflict of
				 *  a predeclaration with a previous one.
				 */
				sym = find_sym(p1->v_str,~(ET_STRUCTDEF+ET_ENUMDEF),0);
				if(sym) {
					if(!typematch(t1,sym->type)) {
						errl(WARN_FN_TYPE,p1,p1->v_str);
						freetype(sym->type);
						sym->type = t1;
					} else {
						freetype(t1);
					}
					if(sym->storclass != TN_GLOBALDEFINED) {
						if(sc == TN_STATIC  && sym->storclass != TN_STATIC)
							errl(WARN_STATIC,p1,p1->v_str);
						sym->storclass = sc;
					}
				} else {
					sym = make_sym(0,p1->v_str,ET_GLOBAL);
					sym->storclass = sc;
					sym->type = t1;
				}
			} else {
				errl(FUNCTIONNOTEXTERN,p1);
				freetype(t1);
			}
		}
		/*
		 *  Gone: function declarations except typedef.
		 *
		 *  An identifier which has been TYPEDEFed may not be redeclared to
		 *  anything else, including an equivalent typedef, in the same scope.
		 */
		else if(sc == TN_TYPEDEF) {
			if(ptr->link3) errl(TYPEDEF_INIT,p1);
			if(!blocklevel && sym) {
				errl(REDECLARATION,p1,p1->v_str);
				freetype(t1);
			} else {
				sym = make_sym(blocklevel,p1->v_str,ET_TYPEDEF);
				sym->type = t1;
			}
		}
		/*
		 *  Gone: all function declarations, and all typedefs.
		 *
		 *  Next:  It is an error to declare anything of type "void", except
		 *         a typedef or a function (both handled earlier).
		 */
		else if(t2->type == TT_BASICTYPE && t2->tu.flags & TF_VOID) {
			errl(VOIDNOTTYPEDEF,p1);
			freetype(t1);
		}
		/*
		 *  Declarations still not handled have both the following:
		 *    - Storage class not TN_TYPEDEF
		 *    - Type not "void" or "function returning ..."
		 */
		else if(blocklevel) {
			/*
			 *  Handle local declarations.  Note that conflict with a previous
			 *  declaration at the same block level has been weeded out.
			 */
			switch(sc) {
				case -1:
				case TN_AUTO:
				case TN_REGISTER:
					siz = size(t1);
					if(!siz)  errl(DECLSIZE,p1); else {
						stackoffset += siz;
						ar = alignreq(t1);
						stackoffset = (stackoffset+(1<<ar)-1) & (~0)<<ar;
					}
					sym = make_sym(blocklevel,p1->v_str,ET_LOCAL);	
					sym->storclass = TN_AUTO;
					sym->type = t1;
					sym->offset = -stackoffset;

					if(ptr->link3) {
						if(ptr->link3->ntype == TN_INIT_EXPR) {
							handle_computed_init(sym,ptr->link3->link1);
							ptr->link3->link1 = 0;
						} else {
							label = labelcnt++;
							ar = alignreq(t1);
							data(1);
							if(ar) printf("\t.align\t%d\n",ar);
							genlabel(label);
							p2 = ptr->link3;
							handle_constant_init(&p2,t1);
							makecopynode(sym,label,ptr->link3->lineno);
						}
					}
					break;
				case TN_STATIC:
					sym = make_sym(blocklevel,p1->v_str,ET_LOCAL);	
					sym->storclass = TN_STATICLOCAL;
					sym->type = t1;
					siz = size(t1);
					ar = alignreq(t1);
					data(1);
					if(ar) printf("\t.align\t%d\n",ar);
					sym->offset = labelcnt;
					if(ptr->link3) {
						genlabel(labelcnt++);
						p2 = ptr->link3;
						handle_constant_init(&p2,t1);
					} else if(!siz)
						errl(DECLSIZE,p1);
					else 
						printf("\t.lcomm\tL%d,%ld\n",labelcnt++,siz);
					break;
				case TN_EXTERN:
					if(ptr->link3) errl(INIT_LOCAL_EXTERN,p1);
					sym = make_sym(blocklevel,p1->v_str,ET_GLOBAL);	
					sym->storclass = sc;
					sym->type = t1;
					break;
				default:
					panic("bad storclass");
			}
		} else {
			/*
			 *  Handle global declarations.  Special case:  An initialized
			 *  external declaration acts as if no storage class was given.
			 */
			if(ptr->link3 && sc == TN_EXTERN) sc = -1;

			switch(sc) {
				case TN_REGISTER:
				case TN_AUTO:
					errl(STORCLASS,p1);
					sc = TN_EXTERN;
					/*
					 *  Fall through, making the declaration an external one
					 *  to reduce later errors.
					 */
				case TN_EXTERN:
					if(sym) {
						if(!typematch(t1,sym->type))
							errl(REDECLARATION,p1,p1->v_str);

						if(sym->storclass == TN_STATIC) {
							errl(WARN_STATIC,p1,p1->v_str);
							sym->storclass = sc;
						}
						freetype(t1);
					} else {
						sym = make_sym(blocklevel,p1->v_str,ET_GLOBAL);	
						sym->storclass = sc;
						sym->type = t1;
					}
					break;
				case TN_STATIC:
				case -1:
					if(sym) {
						if(!typematch(t1,sym->type)) {
							errl(REDECLARATION,p1,p1->v_str);
							break;
						}
						if(sym->storclass == TN_GLOBALDEFINED) {
							if(ptr->link3) errl(REDECLARATION,p1,p1->v_str);
							break;
						}
						freetype(t1);
					} else {
						sym = make_sym(blocklevel,p1->v_str,ET_GLOBAL);	
						sym->storclass = TN_GLOBALDEFINED;
						sym->type = t1;
					}
					ar = alignreq(t1);
					data(1);
					if(ar) printf("\t.align\t%d\n",ar);
					if(ptr->link3) {
						if(sc != TN_STATIC)
							printf("\t.globl\t_%s\n",p1->v_str);
						printf("_%s:\n",p1->v_str);
						p2 = ptr->link3;
						handle_constant_init(&p2,t1);
					} else {
						siz = size(t1);
						if(!siz) errl(DECLSIZE,p1); else {
							printf("\t%s\t_%s,%ld\n",
								(sc == TN_STATIC) ? ".lcomm" : ".comm",
								p1->v_str,siz);
						}
					}
					break;
				default:
					panic("bad storclass");
			}
		}
		ptr = ptr->link1;
	}
	freetype(tnode);
}
