#include "def.h"
#include "error.h"
#include "nodes.h"
static char *sccsid = "@(#) error.c 1.41@(#)";

/*
 *  Table of error messages corresponding to error numbers 1..n
 */
char *errtbl[] = {
	"invalid character (ignored)",
	"unterminated character/string constant",
	"syntax error",
	"preprocessor directive (ignored)",
	"')' expected",
	"';' expected",
	"'while' expected",
	"floating point not supported",
	"can't take size of a function",
	"array of function",
	"function returning function",
	"typedef with initialization",
	"bad storage class on function declaration",
	"array of void",
	"improper use of void",
	"can't initialize local externals",
	"storage class not valid here",
	"size of declarator not known",
	"structure/union '%s' redefined",
	"function returning array",
	"bit fields not supported",
	"enumeration '%s' redefined",
	"enumeration '%s' undefined",
	"array dimension missing",
	"function '%s' given as structure/union member",
	"'%s' is not a function argument",
	"return type of '%s' does not match previous declaration",
	"'%s' redeclared",
	"'%s' declared static and later extern",
	"integer constant required",
	"sizeof returns zero",
	"array of incomplete type",
	"null character constant",
	"extra characters in character constant ignored",
	"nothing to 'break' out of",
	"no loop to 'continue'",
	"duplicate case value",
	"no cases in switch statement",
	"duplicate 'default' in switch statement",
	"'%s' without switch statement",
	"duplicate label",
	"undefined target for goto",
	"'%s' undefined",
	"lvalue required",
	"pointer required",
	"argument of unary '&' must have an address",
	"arithmetic type required",
	"incomplete type",
	"incompatible types",
	"pointer/pointer conversion",
	"pointer/int conversion",
	"enumeration domain mismatch",
	"invalid type for '%s' operator",
	"call to non-function",
	"'%s' is not a structure member",
	"void type illegal in expression",
	"illegal pointer expression",
	"illegal pointer combination",
	"Invalid subscript type",
	"& before array or function: ignored",
	"invalid conditional type",
	"invalid function argument type",
	"extra initializer(s) ignored",
	"string initializer truncated",
	"bad pointer initializer",
	"struct/union not valid here",
	"bison parser error: %s"
};

int n_errs;		/* number of errors found */

/*
 *  This function prints an error or warning message.  Argument w determines
 *  which.  The error number (index into table above) is given as n.
 *  If this is the first error on a given line of source, the error position
 *  is saved so the scanner can display the line and a pointer to the error
 *  position when the line has been completely read.
 */
void err(w,n,a1,a2,a3)
int w,n;
long a1,a2,a3;
{
   if(*filenamebuf) fprintf(stderr,"\"%s\", ",filenamebuf);
   fprintf(stderr,"line %d, %s %d: ", linecnt, w ? "error" : "warning", n );
   fprintf(stderr,errtbl[n-1],a1,a2,a3);
   putc('\n',stderr);
   if(errpos<0) errpos = tokpos;
   if( w ) n_errs++;
}

/*
 *  This function prints errors which occur when the parse tree is traversed.
 *  At this point it is no longer possible to display where the scanner is
 *  in the source, but a tree node, which is supplied as the argument, is
 *  used to display the line number.
 */
void errl(w,n,node,a1,a2,a3)
int w,n;
long a1,a2,a3;
struct treenode *node;
{
   if(*filenamebuf) fprintf(stderr,"\"%s\", ",filenamebuf);
   fprintf(stderr,"near line %d, %s %d: ", node->lineno,
	   w ? "error" : "warning", n );
   fprintf(stderr,errtbl[n-1],a1,a2,a3);
   putc('\n',stderr);
   if( w ) n_errs++;
}

/*
 *  Error output function for the bison parser.  It swallows up "parse error"
 *  messages, since these are *always* accompanied by syntax errors from
 *  the error calls in the grammar, and are not needed to clutter up the output
 *  further.
 */
void yyerror(s)
char *s;
{
	if(strcmp(s,"parse error")) err(BISONERROR,s);
}
