#include "def.h"
#include "nodes.h"
#include "symtab.h"
#include "types.h"
#include "error.h"
#include "nodetypes.h"
static char *sccsid = "@(#) expr.c 1.36@(#)";

	/*
	 * the following file contains various routines used in typechecking
	 * the parse tree as it is created
	 */

	/*
	 * check if node is an lval node (can be assigned to)
	 * returns 1 if true, o if false
	 */

int islval(node)
struct treenode *node;
{
	for(;;) {
		switch(node->ntype) {
			case TN_EXPRESSION:
				node = node->link1;
				break;
			case TN_VALUE:
				if(node->format != NF_LEAF_CONST)
					return 1;
				else
					return 0;
				break;
			case TN_SUBSCRIPT: case TN_VALUEAT: case TN_DOT: case TN_ARROW:
				return 1;
			default:
				return 0;
		}
	}
}

	/*
	 * check if a typechain is an arithmetic type
	 * return 1 if true, 0 if false
	 */

int isarith(type)
struct type *type;
{
	struct type *t1 = skiprefnodes(type);
	if( t1 == 0 ) return 0;
	if(t1->type == TT_ENUMHEADER) return 1;
	else if(t1->type == TT_BASICTYPE && !(t1->tu.flags & TF_VOID)) return 1;
	else return 0;
}

/*
 * The following routine promotes a node to have a 32-bit integer type by
 * inserting an "extend" node ahead of it if necessary.
 * (address of node is passed and modified)
 */

void extend(ptr)
struct treenode **ptr;
{
	struct type *type;

	if( *ptr == 0 ) {
		panic("extend called with a pointer to NULL");
	}

	type = skiprefnodes((*ptr)->vtype);

	/*
	 *  make extend (const) return a long type
	 *  even though it doesn't actually do anything.
	 */

	if((*ptr)->format == NF_LEAF_CONST) {
		type = (*ptr)->vtype;
		(*ptr)->vtype = (struct type *) safealloc(sizeof(struct type));
		(*ptr)->vtype->type = TT_BASICTYPE;
		(*ptr)->vtype->refcount = 1;
		(*ptr)->vtype->tu.flags = TF_BIT32 |
			( type->tu.flags & ~(TF_BIT8+TF_BIT16) );
		freetype(type);
	} else if(type->tu.flags & TF_BIT8) {
		*ptr = make_plain(
			(type->tu.flags & TF_UNSIGNED)?TN_UBYTETOLONG:TN_SBYTETOLONG,
			*ptr,0,0);
		(*ptr)->vtype = (struct type *) safealloc(sizeof(struct type));
		(*ptr)->vtype->type = TT_BASICTYPE;
		(*ptr)->vtype->refcount = 1;
		(*ptr)->vtype->tu.flags = TF_BIT32 |
			( (*ptr)->link1->vtype->tu.flags & ~TF_BIT8 );
	} else if(type->tu.flags & TF_BIT16) {
		*ptr = make_plain(
			(type->tu.flags & TF_UNSIGNED)?TN_UWORDTOLONG:TN_SWORDTOLONG,
			*ptr,0,0);
		(*ptr)->vtype = (struct type *) safealloc(sizeof(struct type));
		(*ptr)->vtype->type = TT_BASICTYPE;
		(*ptr)->vtype->refcount = 1;
		(*ptr)->vtype->tu.flags = TF_BIT32 |
			( (*ptr)->link1->vtype->tu.flags & ~TF_BIT16 );
	} else if(type->tu.flags & TF_BIT32) {
		/* do nothing */
	} else
		panic("invalid call to extend");
}

	/*
	 * create a node to inform code generator to take address of
	 * following expression
	 * (returns start of new node tree)
	 */

struct treenode *make_addressof(node)
struct treenode *node;
{
	struct type *t1 = skiprefnodes(node->vtype);

	/* get the array subtype if an array */

	if( t1->type == TT_ARRAYOF ) t1 = skiprefnodes(t1->tn.next);

	node = make_plain(TN_ADDRESSOF,node,0,0);
	maketypenode(&t1,TT_POINTERTO);
	node->vtype = t1;
	node->vtype->refcount++;
	node->lineno = node->link1->lineno;
	return node;
}

	/*
	 * returns the valid type to return from an arithmetic
	 * operation (taking care of the property of the sign of both 
	 * operands)
	 */

struct type *make_sign( t1, t2 )
struct type *t1, *t2;
{
	short signtype = 0;
	struct type *ret_type = (struct type *)safealloc(sizeof(struct type));

	if( t1 == 0 || t2 == 0 ) panic("null pointers to make_sign");

	ret_type->type = TT_BASICTYPE;
	ret_type->refcount = 1;

	if( (t1->tu.flags&TF_SIZEMASK) == (t2->tu.flags&TF_SIZEMASK) )
		signtype = (t1->tu.flags&TF_UNSIGNED) |
			(t2->tu.flags&TF_UNSIGNED);
	else

	/* signed property is carried over from the largest type */

		signtype = ( (t1->tu.flags&TF_SIZEMASK)>(t2->tu.flags&TF_SIZEMASK) ?
			t1->tu.flags : t2->tu.flags ) & TF_UNSIGNED ;

	ret_type->tu.flags = TF_BIT32|signtype;
	return ret_type;
}
