/* @(#) exprstack.h 1.17@(#)
 *
 *  This file defines the format of an entry in the expression stack.  Each
 *  entry simply has a type, and two 32-bit values.  The following structure
 *  defines how these are stored.
 */
struct es_entry {
	short type;
	union {
		int val1;
		char *str1;
    } u;
	long val2;
};

/*
 *  This is how deep the expression stack can get before the compiler panics.
 *  It should be plenty.
 */
#define ES_STACKDEPTH 20

/*
 *  These are the addressing modes supported.  "x" refers to u.val1, "y" to
 *  val2.  "@" means that the mode can be deferred.
 */
#define ES_REG			1			/* @ Rx			*/
#define ES_CONST		2			/* 	 #x			*/
#define ES_POSTINCR     3 			/* @ (Rx)+		*/
#define ES_PREDECR		4			/*   -(Rx)		*/
#define ES_DREGOFFSET	5			/* @ y(Rx)		*/
#define ES_LABELADDR	6 			/*   #_str1+y	*/
#define ES_STATICADDR  	7 			/*   #Lx+y		*/
#define ES_LABEL		8			/* @ _str1+y	*/
#define ES_STATIC		9			/* @ Lx+y		*/

/*
 *  This value is added to an addressing mode type field to indicate an extra
 *  level of deferring.
 */
#define ES_DEFERRED		16			/* @...			*/

/*
 *  These are special processor registers.
 */
#define REG_SP			14
#define REG_FP			13
#define REG_AP			12
