#include "def.h"
#include "types.h"

char *sccsid = "@(#) freetype.c 1.25@(#)";

/*
 *  This routine deallocates a type expression.  It decrements the reference
 *  count on all type nodes involved in the expression, and deletes those
 *  whose reference count goes to zero, indicating that they are no longer
 *  needed.
 */
void freetype(tnode)
register struct type *tnode;
{
	register struct type *t1;
	/*
	 *  Avoid recursion where it's easy to do...
	 */
	for(;;) {
		if(!tnode) panic("freetype(0)");
		if(!tnode->refcount) {
			panic("freetype: refcount 0");
		}

		if(tnode->refcount--) return;

		switch(tnode->type) {
			case TT_POINTERTO:
			case TT_TYPEREF:
			case TT_ARRAYOF:
			case TT_FUNCTIONRETURNING:
				t1 = tnode;	
				tnode = tnode->tn.next;
				free(t1);
				break;
			case TT_BASICTYPE:
				free(tnode);
				return;
			case TT_STRUCTHEADER:
			case TT_UNIONHEADER:
				{
					struct member *m1, *m = tnode->tn.members;
					while(m) {
						free(m->name);
						freetype(m->type);
						m1 = m;
						m = m->next;
						free(m1);
					}
				}
				free(tnode);
				return;
			case TT_ENUMHEADER:
				{
					struct enumerator *en1, *en = tnode->tn.enumlist;
					while(en) {
						free(en->name);
						en1 = en;
						en = en->next;
						free(en1);
					}
				}
				free(tnode);
				return;
			default:
				panic("freetype: bad type chain");
		}
	}
}
