#include "nodes.h"
#include "symtab.h"
#include "def.h"
#include "error.h"

static char *sccsid = "@(#) goto.c 1.12@(#)";

/*
 *  Process a GOTO label.  The first reference to a label creates a symbol
 *  table entry, at block level 1 (function scope), with a pointer to the tree
 *  node representing the label.  Subsequent references all find this symbol
 *  and end up pointing to the same tree node.  Eventually, the defining
 *  occurrence of the label ("label:" as opposed to "goto label") will be
 *  found and a destination label number assigned and filled into the tree
 *  node.  Then at code generation time, all the nodes referring to this label
 *  have a pointer to the tree node where this label node is stored.  And if
 *  the label is never defined, this is immediately obvious from the fact
 *  that no label number is in the tree node yet.
 *
 *   type:   The value for the "ntype" field of the tree node to create
 *   idnode: The identifier node containing the label name
 *   flag:   Indicates whether the label is defined (1) or just used (0) here.
 */
struct treenode *make_gotolabel(type,idnode,flag)
struct treenode *idnode;
int type,flag;
{
	struct symbol *sym = find_sym(idnode->v_str,ET_GOTOLABEL,0);
	struct treenode *tn=(struct treenode *) safealloc(sizeof(struct treenode));
	if(!sym) {
		sym = make_sym(1,idnode->v_str,ET_GOTOLABEL);
		sym->su.nodeptr = tn;
	}
	tn->v_labelno = 0;
	tn->ntype = type;
	tn->vtype = 0;
	tn->format = NF_LEAF_GOTOLABEL;
	tn->lineno = linecnt;
	tn->v_defnode = sym->su.nodeptr;

	if(flag) {
		if(sym->su.nodeptr->v_labelno)
			err(DUPLABEL);
		else 
			sym->su.nodeptr->v_labelno = labelcnt++;
	}
	kill_tree(idnode);
	return tn;
}
