%{
#include "def.h"
#include "nodetypes.h"
#include "symtab.h"
#include "nodes.h"
#include "error.h"
#include "types.h"

long stackoffset,maxstack;
int blocklevel;
int offset_stack[MAXBLOCKLEVEL];
struct treenode *init_stat_list[MAXBLOCKLEVEL];
extern char stringbuf[];

static char *sccsid = "@(#) grammar.y 1.47@(#)";
%}
/*
 *  The header file "nodetypes.h" is created by scanning this file for
 *  the "TN" constants.  The following will also appear because they are
 *  found in this comment:
 *
 *  TN_GLOBALDEFINED
 *  TN_STATICLOCAL
 *  TN_VALUE
 *  TN_STRING
 *  TN_FUNCTIONCALL
 *  TN_SBYTETOLONG
 *  TN_UBYTETOLONG
 *  TN_SWORDTOLONG
 *  TN_UWORDTOLONG
 *  TN_PTRADD
 *  TN_PTRSUB
 */

%union
{
	struct treenode *node;
	char *string;
	long longint;
}

/*
 * ANSI C grammar
 */


/* Keywords */

%token AUTO BREAK CASE CHAR CONST CONTINUE DEFAULT DO DOUBLE ELSE ENUM
%token EXTERN FLOAT FOR GOTO IF INT LONG REGISTER RETURN SHORT SIGNED
%token SIZEOF STATIC STRUCT SWITCH TYPEDEF UNION UNSIGNED VOID VOLATILE
%token WHILE

/* Constants */

%token <longint> DOUBLENUM FLOATNUM INTEGER LONGINT
%token <longint> UNSLONGINT UNSINT

%token <longint> SINGLECHAR STRING

/* Identifiers */

%token <string> ID TYPE_ID


/* Delimiters */

%token AND			/* & */
%token AND_ASS		/* &= */
%token ARROW		/* -> */
%token ASS			/* = */
%token COLON		/* : */
%token COMMA		/* , */
%token DECR			/* -- */
%token DIV			/* / */
%token DIV_ASS		/* /= */
%token DOT			/* . */
%token DOTDOT		/* .. */
%token EQ			/* == */
%token EXCLAIM		/* ! */
%token EXOR			/* ^ */
%token EXOR_ASS		/* ^= */
%token GE			/* >= */
%token GT			/* > */
%token INCR			/* ++ */
%token LBRACE		/* { */
%token LE			/* <= */
%token LOGAND		/* && */
%token LOGOR		/* || */
%token LPAR			/* ( */
%token LSHIFT		/* << */
%token LSHIFT_ASS	/* <<= */
%token LSQR			/* [ */
%token LT			/* < */
%token MINUS		/* - */
%token MINUS_ASS	/* -= */
%token MOD			/* % */
%token MOD_ASS		/* %= */
%token NE			/* != */
%token NOT			/* ~ */
%token OR			/* | */
%token OR_ASS		/* |= */
%token PLUS			/* + */
%token PLUS_ASS		/* += */
%token QUEST		/* ? */
%token RBRACE		/* } */
%token RPAR			/* ) */
%token RSHIFT		/* >> */
%token RSHIFT_ASS	/* >>= */
%token RSQR			/* ] */
%token SEMI			/* ; */
%token STAR			/* * */
%token STAR_ASS		/* *= */

%type <node> goal identifier type_identifier constant primary_expression
%type <node> postfix_expression argument_list unary_expression cast_expression
%type <node> multiplicative_expression additive_expression shift_expression
%type <node> relational_expression equality_expression exclusive_OR_expression
%type <node> inclusive_OR_expression logical_AND_expression
%type <node> logical_OR_expression conditional_expression assignment_expression
%type <node> expression expression_1 constant_expression declaration
%type <node> declaration_specifiers init_declarator_list init_declarator_list_1
%type <node> init_declarator storage_class_specifier type_specifier
%type <node> struct_or_union struct specifier_qualifier_list
%type <node> struct_declarator any_id enum_specifier enumerator_list
%type <node> enumerator_list_1 enumerator type_qualifier declarator
%type <node> direct_declarator pointer type_qualifier_list parameter_type_list
%type <node> parameter_list parameter_declaration identifier_list type_name
%type <node> abstract_declarator direct_abstract_declarator initializer
%type <node> initializer_list statement labeled_statement label
%type <node> compound_statement declaration_list declaration_list_1
%type <node> statement_list statement_list_1 expression_statement
%type <node> selection_statement iteration_statement jump_statement
%type <node> matched_stat labelled_matched_stat opt_expression translation_unit
%type <node> external_definition function_definition function_header
%type <node> AND_expression struct_declaration_list_1
%type <node> struct_declaration_list struct_declaration struct_declarator_list
%type <node> argument_declarations 

%%

goal
	: translation_unit
	;

identifier
	: ID
		{	$$ = make_str(TN_ID,$1); }
	;

type_identifier
	: TYPE_ID
		{	$$ = make_str(TN_TYPEID,$1); }
	;

constant
	: INTEGER
		{	$$ = make_const($1,0,0); }
	| SINGLECHAR
		{	$$ = make_const(stringbuf[0],0,0);
			if(!$1) err(NULLCHARCONST);
			else if($1>1) err(CHARTOOLONG);
		}
	| LONGINT
		{	$$ = make_const($1,0,1); }
	| UNSINT
		{	$$ = make_const($1,1,0); }
	| UNSLONGINT
		{	$$ = make_const($1,1,1); }
	| FLOATNUM
		{	$$ = make_const($1,0,0); }
	| DOUBLENUM
		{	$$ = make_const($1,0,0); }
	;

primary_expression
	: identifier
		{	$$ = make_var($1); }
	| constant
	| STRING
		{	$$ = make_strconst(stringbuf,$1); }
	| LPAR expression RPAR
		{	$$ = $2; }
	| LPAR expression error
		{	$$ = $2; }
	;

postfix_expression
	: primary_expression
	| postfix_expression LSQR expression RSQR
		{	$$ = make_inner(TN_SUBSCRIPT,$1,$3,NULL); }
	| identifier LPAR argument_list RPAR
		{	$$ = make_functioncall($1,$3); }
	| postfix_expression LPAR argument_list RPAR
		{	$$ = make_functioncall($1,$3); }
	| postfix_expression DOT identifier
		{	$$ = make_inner(TN_DOT,$1,$3,NULL); }
	| postfix_expression ARROW identifier
		{	$$ = make_inner(TN_ARROW,$1,$3,NULL); }
	| postfix_expression INCR
		{	$$ = make_incdec(TN_POSTINCREMENT,$1); }
	| postfix_expression DECR
		{	$$ = make_incdec(TN_POSTDECREMENT,$1); }
	;

argument_list
	:	{	$$ = NULL; }
	| assignment_expression
		{	$$ = make_arg(TN_FUNCTIONARG,NULL,$1); }
	| argument_list COMMA assignment_expression
		{	$$ = make_arg(TN_FUNCTIONARG,$1,$3); }
	;

unary_expression
	: postfix_expression
	| INCR unary_expression
		{	$$ = make_incdec(TN_PREINCREMENT,$2); }
	| DECR unary_expression
		{	$$ = make_incdec(TN_PREDECREMENT,$2); }
	| AND cast_expression
		{	$$ = make_inner(TN_ADDRESSOF,$2,NULL,NULL); }
	| STAR cast_expression
		{	$$ = make_inner(TN_VALUEAT,$2,NULL,NULL); }
	| PLUS cast_expression
		{	$$ = $2; }
	| MINUS cast_expression
		{	$$ = check_arith(TN_UMINUS,$2,NULL,NULL); }
	| NOT cast_expression
		{	$$ = make_inner(TN_BINARYNOT,$2,NULL,NULL); }
	| EXCLAIM cast_expression
		{	$$ = make_cond(TN_LOGICNOT,$2,NULL,NULL); }
	| SIZEOF unary_expression
		{	$$ = eval_sizeof(TN_SIZEOF_EXP,$2); }
	| SIZEOF LPAR type_name RPAR
		{	$$ = eval_sizeof(TN_SIZEOF_TYPE,$3); }
	;

cast_expression
	: unary_expression
	| LPAR type_name RPAR cast_expression
		{	$$ = make_inner(TN_CAST,$2,$4,NULL); }
	;

multiplicative_expression
	: cast_expression
	| multiplicative_expression STAR cast_expression
		{	$$ = check_arith(TN_MULTIPLY,$1,$3,NULL); }
	| multiplicative_expression DIV cast_expression
		{	$$ = check_arith(TN_DIVIDE,$1,$3,NULL); }
	| multiplicative_expression MOD cast_expression
		{	$$ = check_arith(TN_MOD,$1,$3,NULL); }
	;

additive_expression
	: multiplicative_expression
	| additive_expression PLUS multiplicative_expression
		{	$$ = check_arith(TN_ADD,$1,$3,NULL); }
	| additive_expression MINUS multiplicative_expression
		{	$$ = check_arith(TN_SUBTRACT,$1,$3,NULL); }
	;

shift_expression
	: additive_expression
	| shift_expression LSHIFT additive_expression
		{	$$ = check_arith(TN_LSHIFT,$1,$3,NULL); }
	| shift_expression RSHIFT additive_expression
		{	$$ = check_arith(TN_RSHIFT,$1,$3,NULL); }
	;

relational_expression
	: shift_expression
	| relational_expression LT shift_expression
		{	$$ = make_inner(TN_LESSTHAN,$1,$3,NULL); }
	| relational_expression LE shift_expression
		{	$$ = make_inner(TN_LESSEQUAL,$1,$3,NULL); }
	| relational_expression GT shift_expression
		{	$$ = make_inner(TN_GREATERTHAN,$1,$3,NULL); }
	| relational_expression GE shift_expression
		{	$$ = make_inner(TN_GREATEREQUAL,$1,$3,NULL); }
	;

equality_expression
	: relational_expression
	| equality_expression EQ relational_expression
		{	$$ = make_inner(TN_EQUAL,$1,$3,NULL); }
	| equality_expression NE relational_expression
		{	$$ = make_inner(TN_NOTEQUAL,$1,$3,NULL); }
	;

AND_expression
	: equality_expression
	| AND_expression AND equality_expression
		{	$$ = check_arith(TN_BINARYAND,$1,$3,NULL); }
	;

exclusive_OR_expression
	: AND_expression
	| exclusive_OR_expression EXOR AND_expression
		{	$$ = check_arith(TN_EXOR,$1,$3,NULL); }
	;

inclusive_OR_expression
	: exclusive_OR_expression
	| inclusive_OR_expression OR exclusive_OR_expression
		{	$$ = check_arith(TN_BINARYOR,$1,$3,NULL); }
	;

logical_AND_expression
	: inclusive_OR_expression
	| logical_AND_expression LOGAND inclusive_OR_expression
		{	$$ = make_cond(TN_LOGICAND,$1,$3,NULL); }
	;

logical_OR_expression
	: logical_AND_expression
	| logical_OR_expression LOGOR logical_AND_expression
		{	$$ = make_cond(TN_LOGICOR,$1,$3,NULL); }
	;

conditional_expression
	: logical_OR_expression
	| logical_OR_expression QUEST expression COLON conditional_expression
		{	$$ = make_cond(TN_CONDEXP,$1,$3,$5); }
	;

assignment_expression
	: conditional_expression
	| unary_expression ASS assignment_expression
		{	$$ = check_assign(TN_ASSIGN,$1,$3,NULL); }
	| unary_expression STAR_ASS assignment_expression
		{	$$ = check_opassign(TN_MULTASSIGN,$1,$3,NULL); }
	| unary_expression DIV_ASS assignment_expression
		{	$$ = check_opassign(TN_DIVASSIGN,$1,$3,NULL); }
	| unary_expression MOD_ASS assignment_expression
		{	$$ = check_opassign(TN_MODASSIGN,$1,$3,NULL); }
	| unary_expression PLUS_ASS assignment_expression
		{	$$ = check_opassign(TN_ADDASSIGN,$1,$3,NULL); }
	| unary_expression MINUS_ASS assignment_expression
		{	$$ = check_opassign(TN_SUBASSIGN,$1,$3,NULL); }
	| unary_expression LSHIFT_ASS assignment_expression
		{	$$ = check_opassign(TN_LSHIFTASSIGN,$1,$3,NULL); }
	| unary_expression RSHIFT_ASS assignment_expression
		{	$$ = check_opassign(TN_RSHIFTASSIGN,$1,$3,NULL); }
	| unary_expression AND_ASS assignment_expression
		{	$$ = check_opassign(TN_ANDASSIGN,$1,$3,NULL); }
	| unary_expression EXOR_ASS assignment_expression
		{	$$ = check_opassign(TN_EXORASSIGN,$1,$3,NULL); }
	| unary_expression OR_ASS assignment_expression
		{	$$ = check_opassign(TN_ORASSIGN,$1,$3,NULL); }
	;

expression
	: expression_1
		{	if($1->link3 == $1) {
				freetype($1->vtype);
				$$ = $1->link1;
				free($1);
			} else {
				$$ = $1->link3;
				freetype($$->vtype);
				$$->vtype = $1->vtype;
				$1->vtype->refcount++;
				$1->link3 = NULL;
			}
		}
	;

expression_1
	: assignment_expression
		{	$$ = make_inner(TN_EXPRESSION,$1,NULL,NULL);
			$$->link3 = $$; }
	| expression_1 COMMA assignment_expression
		{	$$ = make_inner(TN_EXPRESSION,$3,NULL,$1->link3);
			$1->link3 = $$; }
	;

constant_expression
	: conditional_expression
	;

declaration
	: declaration_specifiers SEMI
		{	$$ = make_plain(TN_DECLARATION,NULL,$1,NULL); }
	| declaration_specifiers init_declarator_list SEMI
		{	$$ = make_plain(TN_DECLARATION,NULL,$1,$2); }
	;

declaration_specifiers
	: storage_class_specifier
		{	$$ = make_plain(TN_STORAGECLASS,$1,NULL,NULL); }
	| type_specifier
		{	$$ = make_plain(TN_TYPESPECIFIER,$1,NULL,NULL); }
	| type_qualifier
		{	$$ = make_plain(TN_TYPEQUALIFIER,$1,NULL,NULL); }
	| declaration_specifiers storage_class_specifier
		{	$$ = make_plain(TN_STORAGECLASS,$2,$1,NULL); }
	| declaration_specifiers type_specifier
		{	$$ = make_plain(TN_TYPESPECIFIER,$2,$1,NULL); }
	| declaration_specifiers type_qualifier
		{	$$ = make_plain(TN_TYPEQUALIFIER,$2,$1,NULL); }
	;

init_declarator_list
	: init_declarator_list_1
		{	$$ = $1->link1; $1->link1 = NULL; }
	;

init_declarator_list_1
	: init_declarator
		{	$1->link1 = $1; }
	| init_declarator_list_1 COMMA init_declarator
		{	$3->link1 = $1->link1; $1->link1 = $3; $$ = $3; }
	;

init_declarator
	: declarator
		{	$$ = make_plain(TN_DECLARATOR,NULL,$1,NULL); }
	| declarator ASS initializer
		{	$$ = make_plain(TN_INITDECLARATOR,NULL,$1,$3); }
	;

storage_class_specifier
	: STATIC
		{	$$ = make_plain(TN_STATIC,0,0,0); }
	| EXTERN
		{	$$ = make_plain(TN_EXTERN,0,0,0); }
	| TYPEDEF
		{	$$ = make_plain(TN_TYPEDEF,0,0,0); }
	| REGISTER
		{	$$ = make_plain(TN_REGISTER,0,0,0); }
	| AUTO
		{	$$ = make_plain(TN_AUTO,0,0,0); }
	;

type_specifier
	: CHAR
		{	$$ = make_plain(TN_CHAR,0,0,0); }
	| SHORT
		{	$$ = make_plain(TN_SHORT,0,0,0); }
	| INT
		{	$$ = make_plain(TN_INT,0,0,0); }
	| LONG
		{	$$ = make_plain(TN_LONG,0,0,0); }
	| UNSIGNED
		{	$$ = make_plain(TN_UNSIGNED,0,0,0); }
	| SIGNED
		{	$$ = make_plain(TN_SIGNED,0,0,0); }
	| FLOAT
		{	$$ = make_plain(TN_INT,0,0,0); err(NOFLOAT); }
	| DOUBLE
		{	$$ = make_plain(TN_INT,0,0,0); err(NOFLOAT); }
	| VOID
		{	$$ = make_plain(TN_VOID,0,0,0); }
	| type_identifier
	| enum_specifier
	| struct_or_union any_id
		{	$$ = make_plain(TN_STRUCT,$1,$2,NULL); }
	| struct
	;

struct_or_union
	: STRUCT
		{	$$ = make_plain(TN_STRUCT,0,0,0); }
	| UNION
		{	$$ = make_plain(TN_UNION,0,0,0); }
	;

struct
	: struct_or_union LBRACE struct_declaration_list RBRACE
		{	$$ = make_plain(TN_STRUCT,$1,NULL,$3); }
	| struct_or_union any_id LBRACE struct_declaration_list RBRACE
		{	$$ = make_plain(TN_STRUCT,$1,$2,$4); }
	;

struct_declaration_list
	: struct_declaration_list_1
		{	$$ = $1->link1; $1->link1 = NULL; }
	;

struct_declaration_list_1
	: struct_declaration
		{	$1->link1 = $1; $$ = $1; }
	| struct_declaration_list_1 struct_declaration
		{	$2->link1 = $1->link1; $1->link1 = $2; $$ = $2; }
	;

struct_declaration
	: specifier_qualifier_list struct_declarator_list SEMI
		{	$$ = make_plain(TN_SDECLARATION,NULL,$1,$2->link1); 
			$2->link1 = NULL; }
		
	;

struct_declarator_list
	: struct_declarator
		{	$1->link1 = $1; $$ = $1; }
	| struct_declarator_list COMMA struct_declarator
		{	$3->link1 = $1->link1; $1->link1 = $3; $$ = $3; }
	;

specifier_qualifier_list
	: type_specifier
		{	$$ = make_plain(TN_TYPESPECIFIER,$1,NULL,NULL); }
	| type_qualifier
		{	$$ = make_plain(TN_TYPEQUALIFIER,$1,NULL,NULL); }
	| specifier_qualifier_list type_specifier
		{	$$ = make_plain(TN_TYPESPECIFIER,$2,$1,NULL); }
	| specifier_qualifier_list type_qualifier
		{	$$ = make_plain(TN_TYPEQUALIFIER,$2,$1,NULL); }
	;

struct_declarator
	: declarator
		{	$$ = make_plain(TN_STRUCTDECL,NULL,$1,NULL); }
	| declarator COLON constant_expression
		{	$$ = make_plain(TN_STRUCTDECL,NULL,$1,$3); }
	| COLON constant_expression
		{	$$ = make_plain(TN_STRUCTDECL,NULL,NULL,$2); }
	;

any_id
	: identifier
	| type_identifier
	;

enum_specifier
	: ENUM LBRACE enumerator_list RBRACE
		{	$$ = make_plain(TN_ENUM,NULL,$3,NULL); }
	| ENUM any_id LBRACE enumerator_list RBRACE
		{	$$ = make_plain(TN_ENUM,$2,$4,NULL); }
	| ENUM any_id
		{	$$ = make_plain(TN_ENUM,$2,NULL,NULL); }
	;

enumerator_list
	: enumerator_list_1
		{	$$ = $1->link1; $1->link1 = NULL; }
	;

enumerator_list_1
	: enumerator
		{ $1->link1 = $1; $$ = $1; }
	| enumerator_list_1 COMMA enumerator
		{ $3->link1 = $1->link1; $1->link1 = $3; $$ = $3; }
	;

enumerator
	: identifier
		{	$$ = make_plain(TN_ENUMERATOR,NULL,$1,NULL); }
	| identifier ASS constant_expression
		{	$$ = make_plain(TN_ENUMERATOR,NULL,$1,$3); }
	;

type_qualifier
	: CONST
		{	$$ = make_plain(TN_CONST,0,0,0); }
	| VOLATILE
		{	$$ = make_plain(TN_VOLATILE,0,0,0); }
	;

declarator
	: direct_declarator
	| pointer direct_declarator
		{	$$ = $1->link1; $1->link1 = $2; }
	;

direct_declarator
	: identifier
	| LPAR declarator RPAR
		{	$$ = $2; }
	| direct_declarator LSQR constant_expression RSQR
		{	$$ = make_plain(TN_SUBSCRIPT,$1,$3,NULL); }
	| direct_declarator LSQR RSQR
		{	$$ = make_plain(TN_SUBSCRIPT,$1,NULL,NULL); }
	| direct_declarator LPAR RPAR
		{	$$ = make_plain(TN_FUNCTIONDECL,$1,NULL,NULL); }
	| direct_declarator LPAR parameter_type_list RPAR
		{	$$ = make_plain(TN_FNDECL_TYPELIST,$1,$3,NULL); }
	| direct_declarator LPAR identifier_list RPAR
		{	$$ = make_plain(TN_FNDECL_IDLIST,$1,$3->link1,NULL);
			$3->link1 = NULL; }
	;

pointer
	: STAR
		{	$$ = make_plain(TN_POINTER,NULL,NULL,NULL);
			$$->link1 = $$; }
	| STAR type_qualifier_list
		{	$$ = make_plain(TN_QUALPOINTER,NULL,$2->link1,NULL);
			$2->link1 = NULL; $$->link1 = $$; }
	| STAR pointer
		{	$$ = make_plain(TN_POINTER,$2->link1,NULL,NULL);
			$2->link1 = $$; $$ = $2; }
	| STAR type_qualifier_list pointer
		{	$$ = make_plain(TN_QUALPOINTER,$3->link1,$2->link1,NULL);
			$2->link1 = NULL; $3->link1 = $$; $$ = $3; }
	;

type_qualifier_list
	: type_qualifier
		{	$$ = make_plain(TN_TYPEQUALLIST,NULL,$1,NULL);
			$$->link1 = $$; }
	| type_qualifier_list type_qualifier
		{	$$ = make_plain(TN_TYPEQUALLIST,$1->link1,$2,NULL);
			$1->link1 = $$; }
	;

parameter_type_list
	: parameter_list
		{	$$ = $1->link1; $1->link1 = NULL; }
	| parameter_list COMMA DOTDOT DOT
		{	$$ = $1->link1;
			$1->link1 = make_plain(TN_DOTDOTDOT,NULL,NULL,NULL); }
	;

parameter_list
	: parameter_declaration
		{	$1->link1 = $1; $$ = $1; }
	| parameter_list COMMA parameter_declaration
		{	$3->link1 = $1->link1; $1->link1 = $3; $$ = $3; }
	;

parameter_declaration
	: declaration_specifiers declarator
		{	$$ = make_plain(TN_PARAMDECL,NULL,$1,$2); }
	| declaration_specifiers
		{	$$ = make_plain(TN_PARAMDECL,NULL,$1,NULL); }
	| declaration_specifiers abstract_declarator
		{	$$ = make_plain(TN_PARAMDECL,NULL,$1,$2); }
	;

identifier_list
	: identifier
		{	$$ = make_plain(TN_IDLIST,NULL,$1,NULL);
			$$->link1 = $$; }
	| identifier_list COMMA identifier
		{	$$ = make_plain(TN_IDLIST,$1->link1,$3,NULL);
			$1->link1 = $$; }
	;

type_name
	: specifier_qualifier_list
		{	$$ = make_plain(TN_TYPENAME,$1,NULL,NULL,NULL); }
	| specifier_qualifier_list abstract_declarator
		{	$$ = make_plain(TN_TYPENAME,$1,$2,NULL,NULL); }
	;

abstract_declarator
	: pointer
		{	$$ = $1->link1; $1->link1 = NULL; }
	| direct_abstract_declarator
	| pointer direct_abstract_declarator
		{	$$ = $1->link1; $1->link1 = $2; }
	;

direct_abstract_declarator
	: LPAR abstract_declarator RPAR
		{	$$ = $2; }
	| LSQR RSQR
        {   $$ = make_plain(TN_SUBSCRIPT,NULL,NULL,NULL); }
	| LSQR constant_expression RSQR
        {   $$ = make_plain(TN_SUBSCRIPT,NULL,$2,NULL); }
	| direct_abstract_declarator LSQR RSQR
        {   $$ = make_plain(TN_SUBSCRIPT,$1,NULL,NULL); }
	| direct_abstract_declarator LSQR constant_expression RSQR
        {   $$ = make_plain(TN_SUBSCRIPT,$1,$3,NULL); }
	| LPAR RPAR
        {   $$ = make_plain(TN_FUNCTIONDECL,NULL,NULL,NULL); }
	| LPAR parameter_type_list RPAR
        {   $$ = make_plain(TN_FNDECL_TYPELIST,NULL,$2,NULL); }
	| direct_abstract_declarator LPAR RPAR
        {   $$ = make_plain(TN_FNDECL_TYPELIST,$1,NULL,NULL); }
	| direct_abstract_declarator LPAR parameter_type_list RPAR
        {   $$ = make_plain(TN_FNDECL_TYPELIST,$1,$3,NULL); }
	;

initializer
	: assignment_expression
		{	$$ = make_plain(TN_INIT_EXPR,$1,NULL,NULL); }
	| LBRACE initializer_list RBRACE
		{	$$ = make_plain(TN_INIT_LIST,$2->link2,NULL,NULL);
			$2->link2 = NULL; }
	| LBRACE initializer_list COMMA RBRACE
		{	$$ = make_plain(TN_INIT_LIST,$2->link2,NULL,NULL);
			$2->link2 = NULL; }
	;

initializer_list
	: initializer
		{ $1->link2 = $1; $$ = $1; }
	| initializer_list COMMA initializer
		{ $3->link2 = $1->link2; $1->link2 = $3; $$ = $3; }
	;

statement
	: labeled_statement
	| compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	| error {	err(ERR_SYNTAX); } SEMI
		{	$$ = make_plain(TN_EXPR_STAT,NULL,NULL,NULL); }
	;

labeled_statement
	: label statement
		{	$$ = make_plain(TN_LABEL,NULL,$1,$2); }
	;

label
	: any_id COLON
		{	$$ = make_gotolabel(TN_GOTOLABEL,$1,1); }
	| CASE constant_expression COLON
		{	struct treenode *ptr = make_plain(TN_CASELABEL,0,0,0);
			$$ = make_plain(TN_CASE,$2,ptr,NULL); }
	| DEFAULT COLON
		{	struct treenode *ptr = make_plain(TN_CASELABEL,0,0,0);
			$$ = make_plain(TN_DEFAULT,ptr,NULL,NULL); }
	;

compound_start
	: LBRACE
		{	if(blocklevel >= MAXBLOCKLEVEL) panic("too many nested blocks");
			offset_stack[blocklevel] = stackoffset;
			init_stat_list[blocklevel] = 0;
			blocklevel++; }
	| compound_start declaration
		{	process_declaration($2);
			kill_tree($2)
		}
	;

compound_end
	: RBRACE
		{	delete_sym_level(blocklevel);
			if(stackoffset > maxstack) maxstack = stackoffset;
			blocklevel--; stackoffset = offset_stack[blocklevel]; }
	;

compound_statement
	: compound_start statement_list compound_end
		{	if(init_stat_list[blocklevel]) {
				$$ = init_stat_list[blocklevel];
				init_stat_list[blocklevel]
					= init_stat_list[blocklevel]->link1;
				$$->link1 = 0;
			}
			$$ = make_plain(TN_COMPOUNDSTAT,NULL,$2,
				init_stat_list[blocklevel]);
		}
	;

declaration_list
	: declaration_list_1
		{	$$ = $1->link1; $1->link1 = NULL; }

declaration_list_1
	: declaration
		{	$1->link1 = $1; $$ = $1; }
	| declaration_list_1 declaration
		{	$2->link1 = $1->link1; $1->link1 = $2; $$ = $2; }
	;

statement_list
	:
		{ 	$$ = NULL; }
	| statement_list_1
		{	$$ = $1->link1; $1->link1 = NULL; }
	;

statement_list_1
	: statement
		{	$1->link1 = $1; $$ = $1; }
	| statement_list_1 statement
		{	$2->link1 = $1->link1; $1->link1 = $2; $$ = $2; }
	;

expression_statement
	: expression SEMI
		{	$$ = make_plain(TN_EXPR_STAT,NULL,$1,NULL); }
	| expression error
		{	err(MISSING_SEMI);
			$$ = make_plain(TN_EXPR_STAT,NULL,$1,NULL); }
	| SEMI
		{	$$ = make_plain(TN_EXPR_STAT,NULL,NULL,NULL); }
	;

selection_statement
	: IF LPAR expression RPAR statement
		{	$$ = make_cond(TN_IF,NULL,$3,$5); }
	| IF LPAR expression RPAR matched_stat ELSE statement
		{	struct treenode *ptr = make_plain(TN_ELSE,$5,$7,NULL);
			$$ = make_cond(TN_IF_ELSE,NULL,$3,ptr); }
	| SWITCH LPAR expression RPAR statement
		{	$$ = make_inner(TN_SWITCH,NULL,$3,$5); }
	;

iteration_statement
	: WHILE LPAR expression RPAR statement
		{	$$ = make_cond(TN_WHILE,NULL,$3,$5); }
	| DO statement WHILE LPAR expression RPAR SEMI
		{	$$ = make_cond(TN_DOWHILE,NULL,$2,$5); }
	| DO statement error
		{	err(WHILE_EXPECTED); $$ = $2; }
	| DO statement WHILE LPAR expression RPAR error
		{	err(MISSING_SEMI);
			$$ = make_cond(TN_DOWHILE,NULL,$2,$5); }
	| FOR LPAR opt_expression SEMI
			opt_expression SEMI
			opt_expression RPAR statement
		{	struct treenode *ptr = make_cond(TN_FOREXP,$3,$5,$7);
			$$ = make_plain(TN_FOR,NULL,ptr,$9); }
	;

jump_statement
	: GOTO any_id SEMI
		{	struct treenode *ptr = make_gotolabel(TN_GOTOLABEL,$2,0);
			$$ = make_plain(TN_GOTO,NULL,ptr,NULL); }
	| GOTO any_id error
		{	struct treenode *ptr = make_gotolabel(TN_GOTOLABEL,$2,0);
			err(MISSING_SEMI); $$ = make_plain(TN_GOTO,NULL,ptr,NULL); }
	| CONTINUE SEMI
		{	$$ = make_plain(TN_CONTINUE,NULL,NULL,NULL); }
	| CONTINUE error
		{	err(MISSING_SEMI); 
			$$ = make_plain(TN_CONTINUE,NULL,NULL,NULL); }
	| BREAK SEMI
		{	$$ = make_plain(TN_BREAK,NULL,NULL,NULL); }
	| BREAK error
		{	err(MISSING_SEMI); $$ = make_plain(TN_BREAK,NULL,NULL,NULL); }
	| RETURN opt_expression SEMI
		{	$$ = make_inner(TN_RETURN,NULL,$2,NULL); }
	| RETURN opt_expression error
		{	err(MISSING_SEMI); $$ = make_plain(TN_RETURN,NULL,$2,NULL); }
	;

matched_stat
	: labelled_matched_stat
	| compound_statement
	| expression_statement
	| jump_statement
	| DO statement WHILE LPAR expression RPAR SEMI
		{	$$ = make_cond(TN_DOWHILE,NULL,$2,$5); }
	| DO statement error
		{	err(WHILE_EXPECTED); $$ = $2; }
	| DO statement WHILE LPAR expression RPAR error
		{	err(MISSING_SEMI);
			$$ = make_cond(TN_DOWHILE,NULL,$2,$5); }
	| IF LPAR expression RPAR matched_stat ELSE matched_stat
		{	struct treenode *ptr = make_plain(TN_ELSE,$5,$7,NULL);
			$$ = make_cond(TN_IF_ELSE,NULL,$3,ptr); }
	| WHILE LPAR expression RPAR matched_stat
		{	$$ = make_cond(TN_WHILE,NULL,$3,$5); }
	| SWITCH LPAR expression RPAR matched_stat
		{	$$ = make_plain(TN_SWITCH,NULL,$3,$5); }
	| FOR LPAR opt_expression SEMI
			opt_expression SEMI
			opt_expression RPAR matched_stat
		{	struct treenode *ptr = make_cond(TN_FOREXP,$3,$5,$7);
			$$ = make_plain(TN_FOR,NULL,ptr,$9); }
	;

labelled_matched_stat
	: label matched_stat
		{	$$ = make_plain(TN_LABEL,NULL,$1,$2); }
	;

opt_expression
	:
		{	$$ = NULL; }
	| expression
	;

translation_unit
	:
	| translation_unit external_definition
	;

external_definition
	: function_definition
	| declaration
		{	blocklevel = 0;
			process_declaration($1);
			kill_tree($1);
		}
	| error
		{	err(ERR_SYNTAX); }
	;

function_definition
	: function_header argument_declarations
	 	{	proc_fnheader($1,$2);
			stackoffset = maxstack = blocklevel = 0;
			kill_tree($1);
			kill_tree($2);
		}
	  compound_statement
		{	reservelocals(); codegen($4,0,0); fn_epilogue(); kill_tree($4); }
	;

function_header
	: declaration_specifiers declarator
		{	$$ = make_plain(TN_FUNCTIONHEADER,$1,$2,NULL); }
	| declarator
		{	$$ = make_plain(TN_FUNCTIONHEADER,NULL,$1,NULL); }
	;

argument_declarations
	:	{	$$ = NULL; }
	| declaration_list
	;
