#include "def.h"
#include "nodes.h"
#include "nodetypes.h"
#include "types.h"
#include "error.h"
#include "symtab.h"
static char *sccsid = "@(#) init.c 1.13@(#)";

extern struct treenode *init_stat_list[];
extern int labelwaiting;
extern int blocklevel;

/*
 *  Generate a statement node to initialize a local variable whose value is
 *  computed at initialization time.  This is passed to the code generator 
 *  before the real statements in the compound statement.
 *
 *  sym  -> symbol table entry for the variable (MUST be local)
 *  node -> parse tree for the expression
 */
handle_computed_init(sym,node)
struct symbol *sym;
struct treenode *node;
{
	struct treenode *vn = (struct treenode *)safealloc(sizeof(struct treenode));
	vn->ntype = TN_VALUE;
	vn->lineno = node->lineno;
	vn->vtype = sym->type;
	sym->type->refcount++;
	vn->format = NF_LEAF_LOCAL;
	vn->v_label = sym->offset;
	node = check_assign(TN_ASSIGN,vn,node,0);
	node = make_plain(TN_EXPR_STAT,0,node,0);
	addtoinitstatlist(node);
}

/*
 *  Save a statement away until after all the declarations are processed, so
 *  its code is only generated after the instruction which reserves space for
 *  the local variables.
 */
addtoinitstatlist(node)
struct treenode *node;
{
	if(!init_stat_list[blocklevel-1])
		node->link1 = node;
	else {
		node->link1 = init_stat_list[blocklevel-1]->link1;
		init_stat_list[blocklevel-1]->link1 = node;
	}
	init_stat_list[blocklevel-1] = node;
}

/*
 *  Handle an aggregate/constant initializer.
 *
 *     node  -> tree node of TN_INIT or TN_INIT_LIST
 *	   tnode -> type expression for the thing to be initialized.
 *
 *  Note that the constant expression stuff is a kludge and cannot handle all
 *  legal intializer constant expressions.  See the documentation for details.
 */
handle_constant_init(cnode,tnode)
struct treenode **cnode;
struct type *tnode;
{
	struct type *t1;
	long val;
	int i;
	struct treenode *node = *cnode;
	int flag;
	int uflag = 0;
	struct member *mp;
	int sp;
	for(;;) {
		switch(tnode->type) {
			case TT_TYPEREF:
				tnode = tnode->tn.next;
				break;
			case TT_BASICTYPE:
				*cnode = node->link2;
				while(node->ntype == TN_INIT_LIST) {
					node = node->link1;
					if(node->link2) errl(EXTRA_INIT_IGNORED,node->link2);
				}
				switch(size(tnode)) {
					case 1:
						printf("\t.byte\t");
						break;
					case 2:
						printf("\t.word\t");
						break;
					default:
						printf("\t.long\t");
				}
				printf("%ld\n",get_const(node->link1));
				labelwaiting = 0;
				return;
			case TT_FUNCTIONRETURNING:
				panic("bad type in handle_constant_init");
			case TT_ENUMHEADER:					
				*cnode = node->link2;
				while(node->ntype == TN_INIT_LIST) {
					node = node->link1;
					if(node->link2) errl(EXTRA_INIT_IGNORED,node->link2);
				}
				t1 = skiprefnodes(node->link1->vtype);
				if(t1->type != TT_ENUMHEADER) {
					errl(ENUMMISMATCH,node->link1);
					val = get_const(node->link1);
				} else if(node->link1->format != NF_LEAF_CONST) {
					errl(NOT_INT_CONSTANT,node->link1);
					val = 0;
				} else
					val = node->link1->v_sval;
				printf("\t.long\t%ld\n",val);
				labelwaiting = 0;
				return;
			case TT_POINTERTO:
				*cnode = node->link2;
				while(node->ntype == TN_INIT_LIST) {
					node = node->link1;
					if(node->link2) errl(EXTRA_INIT_IGNORED,node->link2);
				}
				node = node->link1;
				uflag = 0;
				switch(node->ntype) {
					case TN_STRING:
						data(2);
						genlabel(labelcnt);
						genstring(node,node->v_strlen,1);
						data(1);
						printf("\t.long\tL%d\n",labelcnt++);
						labelwaiting = 0;
						break;
					case TN_ADDRESSOF:
						if(node->link1->ntype != TN_VALUE) goto ptrerr;
						switch(node->link1->format) {
							case NF_LEAF_STATIC:
								printf("\t.long\tL%d\n",node->link1->v_label);
								break;
							case NF_LEAF_LABEL:
								printf("\t.long\t_%s\n",node->link1->v_name);
								break;
							default:
								goto ptrerr;
						}
						break;
					case TN_VALUE:
						switch(node->format) {
							case NF_LEAF_CONST:
								printf("\t.long\t%ld\n",node->v_sval);
								labelwaiting = 0;
								if(!node->v_sval) return;
								break;
							case NF_LEAF_STATIC:
							case NF_LEAF_LABEL:
								t1 = skiprefnodes(node->vtype);
								if(t1->type != TT_ARRAYOF &&
								   t1->type != TT_FUNCTIONRETURNING)
								   goto ptrerr;
								if(node->format == NF_LEAF_STATIC)
									printf("\t.long\tL%d\n",node->v_label);
								else
									printf("\t.long\t_%s\n",node->v_name);
						}
						break;
					default:
ptrerr:
						errl(BADPTRINIT,node);
						printf("\t.long\t0\n");
						labelwaiting = 0;
						return;
				}
				/*
				 *  Well, we've assigned something to this pointer and if
				 *  we are still here, that means we want to check its type
				 *  too.  So do it.
				 */
				t1 = skiprefnodes(node->vtype);
				switch(t1->type) {
					case TT_FUNCTIONRETURNING:
						if(typematch(tnode->tn.next,t1)) return;
						break;
					case TT_ARRAYOF:
						if(typematch(tnode->tn.next,t1->tn.next)) return;
						break;
					case TT_POINTERTO:
						if(typematch(tnode,t1)) return;
						break;
					case TT_BASICTYPE:
						errl(PTRINTCONV,node);
						return;
				}
				errl(PTRPTRCONV,node);
				return;
			case TT_ARRAYOF:
				/*
				 *  Go in one level of {}'s if present.
				 */
				if(node->ntype == TN_INIT_LIST) {
					*cnode = node->link2;
					node = node->link1;
					flag = 0;
				} else
					flag = 1;
				/*
				 *  If there is a single string initializer here, then
				 *  initialize the array with that string.  But only if the
				 *  array elements are 1 byte in size.
				 */
				if(node->ntype==TN_INIT_EXPR && node->link1->ntype==TN_STRING
					&& size(tnode->tn.next) == 1) {
					i = node->link1->v_strlen;
					if(tnode->tu.array.numentries) {
						uflag = 0;
						sp = 0;
						if(tnode->tu.array.numentries < i) {
							errl(INITSTRINGTRUNC,node->link1);
							i = tnode->tu.array.numentries;
						} else if(tnode->tu.array.numentries > i) {
							uflag = 1;
							sp = tnode->tu.array.numentries - i - 1;
						}
					} else {
						tnode->tu.array.numentries = i+1;
						uflag = 1;
						sp = 0;
					}
					genstring(node->link1,i,uflag);
					if(sp) {
						printf("\t.space\t%d\n",sp);
						labelwaiting = 0;
					}
					node = node->link2;
				} else {
					/*
					 *  Now we can just follow node->link2 until we have enough
					 *  items or it becomes null.
					 */
					for(i=0;node;i++) {
						if(tnode->tu.array.numentries &&
						   i >= tnode->tu.array.numentries) break;
						handle_constant_init(&node,tnode->tn.next);
					}
					/*
					 *  Either we ran out of initializers or we have filled all
					 *  elements of the array.  Finish up appropriately.
					 */
					if(!tnode->tu.array.numentries)
						tnode->tu.array.numentries = i;
					else if(tnode->tu.array.numentries > i) {
						printf("\t.space\t%d\n",(tnode->tu.array.numentries-i)
							* tnode->tu.array.sizeeach);
						labelwaiting = 0;
					}
				}
				if(flag) 
					*cnode = node;
				else if(node)
					errl(EXTRA_INIT_IGNORED,node);
				return;
			case TT_UNIONHEADER:
				uflag = 1;
			case TT_STRUCTHEADER:
				if(node->ntype == TN_INIT_LIST) {
					*cnode = node->link2;
					node = node->link1;
					flag = 0;
				} else
					flag = 1;
				/*
				 *  Initialize structure members until no members
				 *  remain.  Pad with zeros after initializers run out.
				 */
				for(mp=tnode->tn.members;mp;mp=mp->next) {
					if(node) {
						handle_constant_init(&node,mp->type);
						sp = size(mp->type);
					} else
						sp = 0;
					if(mp->next && !uflag)
						sp = mp->next->offset - mp->offset - sp;
					else
						sp = tnode->tu.su.size - mp->offset - sp;
					if(sp) {
						printf("\t.space\t%d\n",sp);
						labelwaiting = 0;
					}
					if(uflag) break;
				}
				if(flag) 
					*cnode = node;
				else if(node)
					errl(EXTRA_INIT_IGNORED,node);
				return;
			default:
				panic("bad type chain\n");
		}
	}
}

/*
 *  Make a node to copy a constant initializer into a local variable.
 *
 *    sym   -> symbol table entry for the variable
 *    label -> label number where the initializer is stored
 */
makecopynode(sym,label,lineno)
struct symbol *sym;
int label;
{
	struct treenode *in = (struct treenode *)safealloc(sizeof(struct treenode));
	in->ntype = TN_VALUE;
	in->lineno = lineno;
	in->vtype = sym->type;
	sym->type->refcount++;
	in->format = NF_LEAF_STATIC;
	in->v_label = label;
	handle_computed_init(sym,in);
}
