#include "def.h"
#include <fcntl.h>
#include <stdio.h>
#include <sys/file.h>
#include "error.h"
static char *sccsid = "@(#) main.c 1.36@(#)";

char filenamebuf[80];
int labelcnt;

extern long regmask;
extern int es_pointer;
extern int section;
extern int labelwaiting;
extern int n_errs;

/*
 *  Mainline program for the compiler.  Parses command line options, opens
 *  and closes files, etc.  Calls the parser which drives the rest of the
 *  compiler.
 */
main(argc,argv)
int argc;
char **argv;
{
	extern char *optarg;
	extern int optind;
	short no_cpp = 0,
		to_stdout = 0,
		cookie = 0,
		help = 0;
	char *out_file = 0, c;
	char command[100], tmpbuf[80];

	section = -1;

	/*
	 *  Open input file if given, and remember file name.
	 */

	filenamebuf[0] = 0;
	while( ( c = getopt(argc, argv, "o:icn") ) != EOF ) {
		switch( c ) {
			case 'n': no_cpp++; break;
			case 'c': to_stdout++; break;
			case 'i': cookie++; break;
			case 'o': out_file = optarg; break;
			default: help++;
		}
		if( help ) break;
	}

	if( help ) {
		printf("usage: %s [-n] [-i] [-c] [-o filename] filename\n", argv[0]);
		printf("\t-n             don't use c preprocessor (only for files)\n");
		printf("\t-c             send output to standard out\n");
		printf("\t-i             add version type to output\n");
		printf("\t-o filename    place output into filename\n");
		exit( 1 );
	}

	if( optind != argc ) {	/* take input source from named file */
		int len = strlen( argv[optind] );

		strcpy(tmpbuf,argv[optind]);
		strcpy( filenamebuf, argv[optind]);

		if( len == 1 ) {
			fprintf(stderr,"invalid filename %s\n", tmpbuf);
			exit(1);
		}
		if( tmpbuf[len-1] != 'c' && tmpbuf[len-2] != '.' ) {
			fprintf(stderr,"invalid filename %s\n", tmpbuf);
			exit(1);
		}
		if( access(tmpbuf,R_OK) != 0) {
			fprintf(stderr,"cannot open %s\n", tmpbuf);
			exit(1);
		}
		if( out_file == 0 && to_stdout == 0 ) {
			strcpy( filenamebuf, tmpbuf);
			out_file = tmpbuf;
			tmpbuf[len-1] = 's';
		}
		if( no_cpp ) {		/* don't use c preprocessor */
			yyin = fopen(filenamebuf,"r");
		} else {			/* use c preprocessor */
			sprintf(command,"/lib/cpp %s", filenamebuf);
			yyin = popen(command,"r");
		}
	} else {				/* else take source from stdin */
		no_cpp = 1;
		strcpy( filenamebuf, "(stdin)");
	}
			
	if( out_file ) {		/* send output to file */
		if( freopen(out_file, "w",stdout) == NULL ) {
			fprintf( stderr, "cannot open %s\n", out_file);
			exit( 1 );
		}
	}

	if( cookie ) {
		data(1);
		printf("\t.ascii\t\"@(%c) tcc compiled: %s@(%c)\"\n",
			'#',filenamebuf,'#');
		printf("\t.byte\t0\n");
	}

	/*
	 *  Initialize globals.
	 */

	*linebuf = 0;
	tokpos = 0;
	errpos = -1;
	linecnt = 1;
	labelcnt = 1;
	es_pointer = -1;
	regmask = 0x3f;
	labelwaiting = 0;
	n_errs = 0;
	init_symtab();

	/*
	 *  Go and do all the work.
	 */
	n_errs += yyparse();
	/*
	 *  All done.  Easy, no?
	 */
	if( !no_cpp ) pclose( yyin );
	fclose(stdout);

	if( n_errs && out_file )
		unlink( out_file );

	if( n_errs )
		exit( 1 );
	else
		exit( 0 );
}
