/* @(#) nodes.h 1.42@(#)
 *
 *  This header file defines parse tree nodes.
 *
 *  Nodes can have a variety of formats depending on type.
 */
struct treenode {
	char ntype;					/* values in nodetypes.h */
	char format;				/* see below */
	short lineno;				/* line number in source file */
	struct type *vtype;			/* type chain */
	union {
		struct treenode *links[3]; 		 /* child nodes */
		char *string;

		struct {
			struct treenode *defnode;
			int labelno;
		} gotodata;
		struct {
			struct treenode *link;
			long offset;
		} dotdata;
		struct {
			int length;
			char *ptr;
		} sc;

		long sval;
		unsigned long uval;
		long offset;
		int labelno;
	} tu;
};

/*
 *  Values for the "format" field; these define what the rest of the node
 *  structure stores.  For simplicity, the following macros access the
 *  appropriate fields.
 */

#define NF_INNER			1
#define link1				tu.links[0]
#define link2				tu.links[1]
#define link3				tu.links[2]

#define NF_LEAF_STRING		2
#define v_str				tu.string

#define NF_LEAF_GOTOLABEL	3
#define v_defnode			tu.gotodata.defnode
#define v_labelno			tu.gotodata.labelno

#define NF_LEAF_LOCAL		4
#define v_offset			tu.offset

#define NF_LEAF_STATIC		5
#define v_label				tu.labelno

#define NF_LEAF_LABEL		6
#define v_name				tu.string

#define NF_LEAF_PARAM		7

#define NF_LEAF_CONST		8
#define v_sval				tu.sval
#define v_uval				tu.uval

#define NF_INNER_DOT		9
#define v_dot_link			tu.dotdata.link
#define v_dot_offset		tu.dotdata.offset

#define NF_LEAF_STRCONST	10
#define v_strconst			tu.sc.ptr
#define v_strlen			tu.sc.length

/*
 *  Tree node deletion cleanup:
 *		All:				Free type chain if vtype != 0
 *
 *		NF_INNER:			Delete subtrees hanging off link1-link3
 *		NF_LEAF_STRING:		Free string at v_str
 *		NF_LEAF_LABEL:		Free string at v_name
 *      NF_INNER_DOT		Free subtree hanging off v_dot_link
 *      NF_LEAF_STRCONST:	Free string at v_strlen
 *		others:				Nothing
 *
 *  The following are the external declarations for all the functions which
 *  deal with tree nodes.
 *
 *  Kill tree at rooted at <root>:
 */
extern void kill_tree( /* root */ );
/*
 *  Make an inner tree node of type <type> with subtrees <link1> - <link3>.
 *
 *  This will perform other operations, such as type checking and constant
 *  subexpression elimination, based on <type>.
 */
extern struct treenode *make_inner( /* type, link1, link2, link3 */ );
/*
 *  Make a generic string node.  The string is copied to new storage.
 */
extern struct treenode *make_str( /* type, string */ );
/*
 *  Make a GOTO label leaf node.  <identifier> points to a string node with
 *  the label in it, <definition_flag> is zero if this is a GOTO statement,
 *  nonzero if the label is defined here.
 */
extern struct treenode *make_gotolabel( /* identifier, definition_flag */ );
/*
 *  Evaluate a sizeof().  <node> points to the tree representing the argument
 *  to sizeof.  <type> is TN_SIZEOF_EXP or TN_SIZEOF_TYPE depending on the
 *  type of argument to sizeof().
 */
extern struct treenode *eval_sizeof( /* type, node */ );
/*
 *  Make a node containing a constant of appropriate type.
 */
extern struct treenode *make_const( /* num, unsigned_flag, long_flag */ );
/*
 *  Make a node containing a string constant, type pointer to signed char.
 *  The string is copied to new storage.
 */
extern struct treenode *make_strconst( /* string */ );
/*
 *  Make a node containing a variable.
 */
extern struct treenode *make_var( /* identifier node */ );

extern struct treenode *check_cond( /* type,l1,l2,l3 */ );
extern struct treenode *check_const( /* type,l1,l2,l3 */ );
extern struct treenode *check_arith( /* type,l1,l2,l3 */ );
extern struct treenode *check_assign( /* type,l1,l2,l3 */ );
extern struct treenode *check_opassign( /* type,l1,l2,l3 */ );
extern struct treenode *eval_const( /* type,l1,l2,l3 */ );
extern struct treenode *eval_sizeof( /* type,node */ );
extern struct treenode *make_addressof( /* node */ );
extern struct treenode *make_const( /* num,uflag,lflag */ );
extern struct treenode *make_functioncall( /* node1,node2 */ );
extern struct treenode *make_gotolabel( /* type,idnode,flag */ );
extern struct treenode *make_inner( /* type,l1,l2,l3 */ );
extern struct treenode *make_plain( /* type,l1,l2,l3 */ );
extern struct treenode *make_str( /* type,string */ );
extern struct treenode *make_strconst( /* string */ );
extern struct treenode *make_var( /* id */ );
extern struct treenode *make_cond( /* type, l1, l2, l3 */ );
extern struct treenode *make_arg( /* type,l1,l2 */ );
extern struct treenode *make_incdec( /* type, l1 */ );
