%{
#include "def.h"
#include "error.h"
#include "nodes.h"
#include "grammar.tab.h"
#include "symtab.h"
static char *sccsid = "@(#) scan.l 1.29@(#)";

extern YYSTYPE yylval;

int linecnt, tokpos, errpos;

#define MAXLINE 1000
#define MAXSTR 1000

char linebuf[MAXLINE];
char stringbuf[MAXSTR];
%}

%%
/*
 *  Keywords
 */

auto		{	save();return(AUTO); }
break		{	save();return(BREAK); }
case		{	save();return(CASE); }
char		{	save();return(CHAR); }
const		{	save();return(CONST); }
continue	{	save();return(CONTINUE); }
default		{	save();return(DEFAULT); }
do			{	save();return(DO); }
double		{	save();return(DOUBLE); }
else		{	save();return(ELSE); }
enum		{	save();return(ENUM); }
extern		{	save();return(EXTERN); }
float		{	save();return(FLOAT); }
for			{	save();return(FOR); }
goto		{	save();return(GOTO); }
if			{	save();return(IF); }
int			{	save();return(INT); }
long		{	save();return(LONG); }
register	{	save();return(REGISTER); }
return		{	save();return(RETURN); }
short		{	save();return(SHORT); }
signed		{	save();return(SIGNED); }
sizeof		{	save();return(SIZEOF); }
static		{	save();return(STATIC); }
struct		{	save();return(STRUCT); }
switch		{	save();return(SWITCH); }
typedef		{	save();return(TYPEDEF); }
union		{	save();return(UNION); }
unsigned	{	save();return(UNSIGNED); }
void		{	save();return(VOID); }
volatile	{	save();return(VOLATILE); }
while		{	save();return(WHILE); }

/*
 *  Other tokens
 */

"&"		{	save();return(AND); }
"&="	{	save();return(AND_ASS); }
"->"	{	save();return(ARROW); }
"="		{	save();return(ASS); }
":"		{	save();return(COLON); }
","		{	save();return(COMMA); }
"--"	{	save();return(DECR); }
"/"		{	save();return(DIV); }
"/="	{	save();return(DIV_ASS); }
"."		{	save();return(DOT); }
".."	{	save();return(DOTDOT); }
"=="	{	save();return(EQ); }
"!"		{	save();return(EXCLAIM); }
"^"		{	save();return(EXOR); }
"^="	{	save();return(EXOR_ASS); }
">="	{	save();return(GE); }
">"		{	save();return(GT); }
"++"	{	save();return(INCR); }
"{"		{	save();return(LBRACE); }
"<="	{	save();return(LE); }
"&&"	{	save();return(LOGAND); }
"||"	{	save();return(LOGOR); }
"("		{	save();return(LPAR); }
"<<"	{	save();return(LSHIFT); }
"<<="	{	save();return(LSHIFT_ASS); }
"["		{	save();return(LSQR); }
"<"		{	save();return(LT); }
"-"		{	save();return(MINUS); }
"-="	{	save();return(MINUS_ASS); }
"%"		{	save();return(MOD); }
"%="	{	save();return(MOD_ASS); }
"!="	{	save();return(NE); }
"~"		{	save();return(NOT); }
"|"		{	save();return(OR); }
"|="	{	save();return(OR_ASS); }
"+"		{	save();return(PLUS); }
"+="	{	save();return(PLUS_ASS); }
"?"		{	save();return(QUEST); }
"}"		{	save();return(RBRACE); }
")"		{	save();return(RPAR); }
">>"	{	save();return(RSHIFT); }
">>="	{	save();return(RSHIFT_ASS); }
"]"		{	save();return(RSQR); }
";"		{	save();return(SEMI); }
"*"		{	save();return(STAR); }
"*="	{	save();return(STAR_ASS); }

/*
 *  Integer constants, in three numbering systems (hex, octal, and decimal)
 *  and with optional postfixes L, U, UL (case independent) to indicate
 *  type.
 */

0[xX][a-fA-F0-9]+		{	save(); sscanf(yytext+2,"%lx",&yylval.longint);
							return(INTEGER); }
0[xX][a-fA-F0-9]+[uU]	{	save(); sscanf(yytext+2,"%lx",&yylval.longint);
							return(UNSINT); }
0[xX][a-fA-F0-9]+[lL]	{	save(); sscanf(yytext+2,"%lx",&yylval.longint);
							return(LONGINT); }
0[xX][a-fA-F0-9]+[uU][lL]	{	save(); sscanf(yytext+2,"%lx",&yylval.longint);
								return(UNSLONGINT); }
0[0-7]+					{	save(); sscanf(yytext+1,"%lo",&yylval.longint);
							return(INTEGER); }
0[0-7]+[uU]				{	save(); sscanf(yytext+1,"%lo",&yylval.longint);
							return(UNSINT); }
0[0-7]+[lL]				{	save(); sscanf(yytext+1,"%lo",&yylval.longint);
							return(LONGINT); }
0[0-7]+[uU][lL]			{	save(); sscanf(yytext+1,"%lo",&yylval.longint);
							return(UNSLONGINT); }
[0-9]+					{	save(); sscanf(yytext,"%ld",&yylval.longint);
							return(INTEGER); }
[0-9]+[uU]				{	save(); sscanf(yytext,"%ld",&yylval.longint);
							return(UNSINT); }
[0-9]+[lL]				{	save(); sscanf(yytext,"%ld",&yylval.longint);
							return(LONGINT); }
[0-9]+[uU][lL]			{	save(); sscanf(yytext,"%ld",&yylval.longint);
							return(UNSLONGINT); }

/*
 *  Floating point constants.  Enforce the rule that at least one
 *  of integer part, fraction part, or exponent must be present.
 */

[0-9]+\.[0-9]*([eE][+-]?[0-9]+)?[fFlL]?		{	save(); return(FLOATNUM); }
[0-9]*\.[0-9]+([eE][+-]?[0-9]+)?[fFlL]?		{	save(); return(FLOATNUM); }
[0-9]+[eE][+-]?[0-9]+[fFlL]?				{	save(); return(FLOATNUM); }

/*
 *  String and character constants.
 */

'			{	save(); dostring('\''); return(SINGLECHAR); }
\"			{	save(); dostring('\"'); return(STRING); }

/*
 *  Preprocessor directives.  Interpret those which indicate where we are in
 *  what source file, throw all others away.
 */

^#[ \t]*line.+\n	sscanf(yytext+1,"%*s %d \"%[^\"]",&linecnt,filenamebuf);
^#[ \t]*[0-9]+.*\n	sscanf(yytext+1,"%d \"%[^\"]",&linecnt,filenamebuf);
^#.*$				{	save(); err(WARN_PREPROC); }

/*
 *  Comments.  Eat these in case the preprocessor didn't.  A subroutine is
 *  used rather than providing an expression for the comment, since comments
 *  can get very long and could overflow the lex input buffer easily otherwise.
 */

"/*"				{	save(); eat_comment(); }

/*
 *  Identifiers
 */

[a-zA-Z_][a-zA-Z_0-9]*	{	save(); return handle_id(); }

/*
 *  White space
 */

[ \t\n\f\r\v\a]			save();

/*
 *  Anything else
 */
.			{	save();err(WARN_INVALID_CHAR); }
%%

/*
 *  This subroutine skips the rest of a comment after the initial "/*" has
 *  been recognized.  End of file is considered to be end of comment.
 */

eat_comment()
{
	int c;
	for(;;) {
		do {
			c = igetc();
			if(c<=0) return;
		} while(c!='*');
		c = igetc();
		if(c<=0 || c=='/') return;
	}
}

int handle_id()
{
	struct symbol *sym = find_sym(yytext,~(ET_STRUCTDEF+ET_ENUMDEF),0);
	yylval.string = save_str(yytext);

	if(sym) if(sym->entrytype == ET_TYPEDEF) return TYPE_ID;
	return ID;
}

/*
 *  This subroutine is a major kludge.  It reconstructs the input lines to
 *  (f)lex by storing everything that is scanned.  It only stores the current
 *  input line.
 *
 *  When a line is complete, the routine checks if an error occurred on it,
 *  and if so, prints it with a pointer to the error.
 */

save()
{
	int i;
	char *p1, *p2;
	p1 = p2 = yytext;
	while(*p1) if(*p1++ == '\n') {
		if(errpos>=0) {
			strncat(linebuf,yytext,p1-yytext);
			fprintf(stderr,"%s",linebuf);
			for(i=0;i<errpos;i++)
				putc(linebuf[i]=='\t' ? '\t' : ' ', stderr);
			fprintf(stderr,"^\n");
			errpos = -1;
		}
		p2 = p1;
		linecnt++;
		*linebuf = 0;
		break;
	}
	tokpos = strlen(linebuf);
	if(tokpos+strlen(p2) >= MAXLINE) {
		fprintf(stderr,"*** Fatal: Input line too long.\n");
		exit(99);
	} else 
		strcpy(linebuf+tokpos,p2);
}

/*
 *  Major kludge part two.  This subroutine obtains the next character
 *  via input(), and also passes it to save().  This must be called instead
 *  of input().
 */

int igetc()
{
	yytext[0] = input();
	yytext[1] = 0;
	save();
	return yytext[0];
}

/*
 *  This subroutine handles string and character constants.
 */

dostring(delim)
char delim;
{
	char *p = stringbuf;
	int c1;
	int c = igetc();
	
	for(;;) {
		if(c==delim) break;
		else if(c=='\n' || c<=0) {
			err(ERR_UNTERMINATED_CHAR_OR_STRING);
			break;
		} else if(c=='\\') {
			c = igetc();
			if(c<=0) {
				err(ERR_UNTERMINATED_CHAR_OR_STRING);
				break;
			}
			switch(c) {
				case 'n': *p++ = '\n'; break;
				case 't': *p++ = '\t'; break;
				case 'v': *p++ = '\v'; break;
				case 'b': *p++ = '\b'; break;
				case 'r': *p++ = '\r'; break;
				case 'f': *p++ = '\f'; break;
				case 'a': *p++ = '\a'; break;
				case 'x':
				case 'X':
					c1 = 0;
					for(;;) {
						c = igetc();
						if(c>='0' && c<='9') c1 = (c1<<4) + c - '0';
						else if(c>='A' && c<='F') c1 = (c1<<4) + c - ('A'-10);
						else if(c>='a' && c<='f') c1 = (c1<<4) + c - ('a'-10);
						else break;
					}
					*p++ = c1;
					continue;
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
					c1 = 0;
					for(;;) {
						if(c>='0' && c<='7') c1 = (c1<<3) + c - '0';
						else break;
						c = igetc();
					}
					*p++ = c1;
					continue;
				case '\n': break;
				default: *p++ = c;
			}
		} else {
			*p++ = c;
		}
		c = igetc();
	}
	yylval.longint = p-stringbuf;
	return;
}
