#include "def.h"
#include "types.h"
#include "error.h"
#include "symtab.h"
static char *sccsid = "@(#) size.c 1.22@(#)";

/*
 *  Figure out the size of a type given a type chain for it.  This routine
 *  is written non-recursively for speed.
 */
long size(node)
register struct type *node;
{
	register int siz = 1;
	register struct symbol *sym;

    for(;;) switch(node->type) {
		case TT_POINTERTO:
			return 4;
		case TT_FUNCTIONRETURNING:
			return 0;
		case TT_ARRAYOF:
			return siz * node->tu.array.numentries
					   * node->tu.array.sizeeach;
		case TT_TYPEREF:
			node = node->tn.next;
			break;
		case TT_STRUCTHEADER:
		case TT_UNIONHEADER:
			return node->tu.su.size;
		case TT_ENUMHEADER:
			return 4;
		case TT_BASICTYPE:
			if(node->tu.flags & TF_BIT32) return 4;
			else if(node->tu.flags & TF_BIT16) return 2;
			else if(node->tu.flags & TF_BIT8) return 1;
			else return 0;
		default:
			panic("bad type chain");
	}
}

/*
 *  Compute the alignment requirement of a type.  The alignment requirement
 *  is returned as a number of bits, suitable for the ".align" op in the VAX
 *  assembler.
 *
 *  pointers, enums, long ints:  on a longword boundary
 *  short ints:				     on a word boundary
 *	arrays:						 alignment requirements of members
 *  structs/unions:				 alignment requirement computed when the
 *								 structure was defined.
 */
int alignreq(tnode)
struct type *tnode;
{
	for(;;) switch(tnode->type) {
		case TT_POINTERTO:
		case TT_ENUMHEADER:
			return 2;
		case TT_FUNCTIONRETURNING:
			return 1;
		case TT_ARRAYOF:
		case TT_TYPEREF:
			tnode = tnode->tn.next;
			break;
		case TT_STRUCTHEADER:
		case TT_UNIONHEADER:
			return tnode->tu.su.alignreq;
		case TT_BASICTYPE:
			if(tnode->tu.flags & TF_BIT32) return 2;
			else if(tnode->tu.flags & TF_BIT16) return 1;
			else return 0;
		default:
			panic("bad type chain");
	}
}
