#include "def.h"
#include "symtab.h"
static char *sccsid = "@(#) symtab.c 1.30@(#)";

struct symbol *symtab_head;

/*
 *  Initialize the symbol table to empty.  This was put into a function so
 *  that the symbol table manager can be a separate "module", accessed
 *  through functions rather than direct manipulation of its variables.
 */
void init_symtab()
{
	symtab_head = 0;
}

/*
 *  Create a symbol table node.  The node is linked into the queue just
 *  ahead of the newest node of the same or lower block level.
 *
 *   level:     the block level of this symbol (0 if global)
 *   name:      name of the symbol
 *   entrytype: ET_GLOBAL, ET_LOCAL, etc... (see symtab.h)
 */
struct symbol *make_sym(level,name,entrytype)
register int level;
int entrytype;
char *name;
{
	register struct symbol **ptr = &symtab_head;

	struct symbol *s = (struct symbol *)
		safealloc(sizeof(struct symbol) + strlen(name));
	s->entrytype = entrytype;
	s->blocklevel = level;
	s->type = 0;
	strcpy(s->name,name);

	while(*ptr && (*ptr)->blocklevel > level) ptr = &(*ptr)->next;

	s->next = *ptr;
	*ptr = s;

	return s;
}

/*
 *  Pop the indicated level(s) of symbols off the symbol table, freeing the
 *  nodes and their type chains.
 */
void delete_sym_level(level)
register int level;
{
	register struct symbol *h;
	while(symtab_head && symtab_head->blocklevel >= level) {
		h = symtab_head;
		symtab_head = symtab_head->next;
		if(h->type) freetype(h->type);
		free(h);
	}
}

/*
 *  Find the next symbol matching the name and entry type, starting at a
 *  current symbol node.
 *
 *	 ptr:       the symbol node to start searching at.
 *   name:      the name to look for or null if any symbol of a given type.
 *   entrytype:	a bit mask of entry types to look for
 *				 (e.g. ET_STRUCTDEF | ET_ENUMDEF)
 *   minlevel:  the minimum acceptable block level.
 */
struct symbol *find_next(ptr,name,entrytype,minlevel)
register struct symbol *ptr;
register char *name;
register int entrytype,minlevel;
{
	while(ptr) {
		if(ptr->blocklevel < minlevel) return 0;
		if(entrytype & ptr->entrytype) 
			if(!name || !strcmp(name,ptr->name)) return ptr;
		ptr = ptr->next;
	}
	return 0;
}

/*
 *  As above, but starts at the head of the symbol table to find the first
 *  eligible symbol.
 */
struct symbol *find_sym(name,entrytype,minlevel)
char *name;
int entrytype,minlevel;
{
	return find_next(symtab_head,name,entrytype,minlevel);
}
