#include "def.h"
#include "types.h"
#include "nodes.h"
#include "nodetypes.h"
#include "error.h"
static char *sccsid = "@(#) typecheck.c 1.31@(#)";

	/*
	 * preform arithmetic typechecking while creating parse tree
	 */

struct treenode *check_arith(type,l1,l2,l3)
int type;
struct treenode *l1, *l2, *l3;
{
	extern struct type *make_int();
	struct type *t1, *t2, *t3;
	struct treenode *tmp, *node = check_const(type,l1,l2,l3);

	/* if reduced to a constant return */

	if( node->format == NF_LEAF_CONST ) return node;

	t1 = skiprefnodes( l1->vtype );
	if( l2 == 0 ) {
		if( isvoid(t1) ) {
			err(VOIDEXPR);
			kill_tree(node);
			return make_const(1,1,1);
		}
		node->vtype = node->link1->vtype;
		node->vtype->refcount++;
		return node;
	}

	/* from now on we are working on two operand operations */

	t2 = skiprefnodes( l2->vtype );

	if( isvoid(t2) ) {
		err(VOIDEXPR);
		kill_tree(node);
		return make_const(1,1,1);
	}

	/* first compute easy stuff (basic types) */

	if( t1->type == TT_BASICTYPE && t2->type == TT_BASICTYPE ) {
		node->vtype = make_sign( t1, t2 );
		if( !( (t1->tu.flags & TF_BIT32) && (t2->tu.flags & TF_BIT32) ) ) {
			extend( &(node->link1) );
			extend( &(node->link2) );
		}
		return node;
	}

	switch( CROSS(t1->type,t2->type,8) ) {
		case CROSS(TT_ENUMHEADER,TT_BASICTYPE,8):
		case CROSS(TT_BASICTYPE,TT_ENUMHEADER,8):
			if( t1->type == TT_BASICTYPE )
				node->vtype = t1;
			else
				node->vtype = t2;
			node->vtype->refcount++;
			break;
		case CROSS(TT_ENUMHEADER,TT_ENUMHEADER,8):
			node->vtype = make_int();
			return node;
	}

	/* check for operations that are valid only on integers */

	if( type == TN_LSHIFT || type == TN_RSHIFT || type == TN_BINARYAND ||
			type == TN_BINARYOR || type == TN_EXOR ||
			type == TN_MULTIPLY || type == TN_DIVIDE ) {
		err(INVALIDTYPEFOROP, 
			type == TN_MULTIPLY		? "*" :
			type == TN_DIVIDE		? "/" :
			type == TN_LSHIFT		? ">>" :
			type == TN_RSHIFT		? "<<" :
			type == TN_BINARYAND	? "&" :
			type == TN_BINARYOR		? "|" :
			type == TN_EXOR			? "^" : "error" );
		kill_tree( node );
		return make_const(1,1,1);
	}

	/* now the complex types */

	switch( CROSS(t1->type,t2->type,8) ) {
		case CROSS(TT_ENUMHEADER,TT_ARRAYOF,8):
		case CROSS(TT_BASICTYPE,TT_ARRAYOF,8):
		case CROSS(TT_BASICTYPE,TT_POINTERTO,8):
			if( type != TN_ADD ) {
				err(INVALIDPOINTEREXPR);
				kill_tree(node);
				return make_const(1,1,1);
			}
			tmp = l1; l1 = l2; l2 = tmp;
			tmp = node->link1; node->link1 = node->link2; node->link2 = tmp;
			t3 = t2; t2 = t1; t1 = t3;
		case CROSS(TT_POINTERTO,TT_ENUMHEADER,8):
		case CROSS(TT_POINTERTO,TT_BASICTYPE,8):
		case CROSS(TT_ARRAYOF,TT_BASICTYPE,8):
			if( t1->type == TT_ARRAYOF )
				node->link1 = l1 = make_addressof( l1 );
			switch( type ) {
				case TN_ADD :
					node->ntype = TN_PTRADD;
					node->vtype = node->link1->vtype;
					node->vtype->refcount++;
					break;
				case TN_SUBTRACT :
					node->ntype = TN_PTRSUB;
					node->vtype = node->link1->vtype;
					node->vtype->refcount++;
					break;
				default:
					err(INVALIDPOINTEREXPR);
					kill_tree(node);
					return make_const(1,1,1);
			}
			break;
		case CROSS(TT_ARRAYOF,TT_POINTERTO,8):
		case CROSS(TT_POINTERTO,TT_ARRAYOF,8):
		case CROSS(TT_POINTERTO,TT_POINTERTO,8):
		case CROSS(TT_ARRAYOF,TT_ARRAYOF,8):

			/* change any references to arrays to pointers */

			if(t1->type == TT_ARRAYOF) node->link1 = l1 = make_addressof( l1 );
			if(t2->type == TT_ARRAYOF) node->link2 = l2 = make_addressof( l2 );

			/* check for equality in pointers */

			if( !typematch( l1->vtype, l2->vtype ) ) {
				err(INVALIDPOINTERCOMB);
				kill_tree(node);
				return make_const(1,1,1);
			}
			switch( type ) {
				case TN_SUBTRACT :
					node->ntype = TN_PTRSUB;
					node->vtype = make_int();
					break;
				default: /* invalid operation for pointer */
					err(INVALIDPOINTEREXPR);
					kill_tree(node);
					return make_const(1,1,1);
			}
			break;
		default:
			err(INCOMPTYPES);
			kill_tree(node);
			return make_const(1,1,1);
	}
	return node;
}

	/*
	 * check for constant subexpressions (if so resolve them)
	 * creates a plain node that is passed on if no subexpressions
	 * can be resolved
	 */

struct treenode *check_const(type,l1,l2,l3)
int type;
struct treenode *l1, *l2, *l3;
{
	struct treenode *tmp, *node = make_plain(type,l1,l2,l3);

	/* check for single operand operators on constant nodes */

	if( l2 == 0 ) {
		if( l1->format == NF_LEAF_CONST ) {
			free( node );
			return eval_const(type,l1,l2,l3);
		}
		return node;
	}

	/*
	 * check for double operand operators on constant nodes
	 * (this is the only check we need to do as any brackets are
	 * removed by the grammar)
	 */

	if( l1->format == NF_LEAF_CONST && l2->format == NF_LEAF_CONST ) {
		free( node );
		return eval_const(type,l1,l2,l3);
	}
	return node;
}

	/*
	 * given a constant expression evalute it and return result 
	 * as a constant node (any extra node freed up)
	 */

struct treenode *eval_const(type,l1,l2,l3)
int type;
struct treenode *l1, *l2, *l3;
{

	/* check for single operand operations */

	if( l2 == 0 ) {
		if( (l1->vtype->tu.flags&TF_UNSIGNED)==TF_UNSIGNED ) { /* unsigned */
			switch( type ) {
				case TN_BINARYNOT :	l1->v_uval = ~ l1->v_uval ; break;
				case TN_LOGICNOT :	l1->v_uval = ! l1->v_uval ; break;
				case TN_UMINUS :	l1->v_uval = - l1->v_uval ; break;
			}
		} else { /* signed */
			switch( type ) {
				case TN_BINARYNOT :	l1->v_sval = ~ l1->v_sval ; break;
				case TN_LOGICNOT :	l1->v_sval = ! l1->v_sval ; break;
				case TN_UMINUS :	l1->v_sval = - l1->v_sval ; break;
			}
		}
		return l1 ;
	}
		
	/* a gross way of making a cross product to typecheck for signed */

	switch( (l1->vtype->tu.flags&TF_UNSIGNED)==TF_UNSIGNED +
		(((l2->vtype->tu.flags&TF_UNSIGNED)==TF_UNSIGNED)<<1)) {

		case 0: /* signed op signed */
			switch( type ) {
				case TN_ADD :		l1->v_sval = l1->v_sval + l2->v_sval;break;
				case TN_BINARYAND :	l1->v_sval = l1->v_sval & l2->v_sval;break;
				case TN_BINARYOR :	l1->v_sval = l1->v_sval | l2->v_sval;break;
				case TN_DIVIDE :	l1->v_sval = l1->v_sval / l2->v_sval;break;
				case TN_EXOR :		l1->v_sval = l1->v_sval ^ l2->v_sval;break;
				case TN_EQUAL :		l1->v_sval = l1->v_sval == l2->v_sval;break;
				case TN_GREATERTHAN:l1->v_sval = l1->v_sval > l2->v_sval;break;
				case TN_GREATEREQUAL:l1->v_sval = l1->v_sval>= l2->v_sval;break;
				case TN_LESSTHAN :	l1->v_sval = l1->v_sval < l2->v_sval;break;
				case TN_LESSEQUAL :	l1->v_sval = l1->v_sval <= l2->v_sval;break;
				case TN_LOGICAND :	l1->v_sval = l1->v_sval && l2->v_sval;break;
				case TN_LOGICOR :	l1->v_sval = l1->v_sval || l2->v_sval;break;
				case TN_LSHIFT :	l1->v_sval = l1->v_sval << l2->v_sval;break;
				case TN_MOD :		l1->v_sval = l1->v_sval % l2->v_sval;break;
				case TN_MULTIPLY :	l1->v_sval = l1->v_sval * l2->v_sval;break;
				case TN_NOTEQUAL :	l1->v_sval = l1->v_sval != l2->v_sval;break;
				case TN_RSHIFT :	l1->v_sval = l1->v_sval >> l2->v_sval;break;
				case TN_SUBTRACT :	l1->v_sval = l1->v_sval - l2->v_sval;break;
				default:
					panic("invalid op for eval const");
			}
			break;
		case 1: /* unsigned op signed */
			l2->v_uval = (unsigned long) l2->v_sval;
			goto evalute_unsigned;
		case 2: /* signed op unsigned */
			l1->v_uval = (unsigned long) l1->v_sval;
			l1->vtype->tu.flags |= TF_UNSIGNED;
		case 3: /* unsigned op unsigned */
evalute_unsigned:
			switch( type ) {
				case TN_ADD :		l1->v_uval = l1->v_uval + l2->v_uval;break;
				case TN_BINARYAND :	l1->v_uval = l1->v_uval & l2->v_uval;break;
				case TN_BINARYOR :	l1->v_uval = l1->v_uval | l2->v_uval;break;
				case TN_DIVIDE :	l1->v_uval = l1->v_uval / l2->v_uval;break;
				case TN_EXOR :		l1->v_uval = l1->v_uval ^ l2->v_uval;break;
				case TN_EQUAL :		l1->v_sval = l1->v_uval == l2->v_uval;break;
				case TN_GREATERTHAN:l1->v_uval = l1->v_uval > l2->v_uval;break;
				case TN_GREATEREQUAL:l1->v_uval = l1->v_uval>= l2->v_uval;break;
				case TN_LESSTHAN :	l1->v_uval = l1->v_uval < l2->v_uval;break;
				case TN_LESSEQUAL :	l1->v_uval = l1->v_uval <= l2->v_uval;break;
				case TN_LOGICAND :	l1->v_uval = l1->v_uval && l2->v_uval;break;
				case TN_LOGICOR :	l1->v_uval = l1->v_uval || l2->v_uval;break;
				case TN_LSHIFT :	l1->v_uval = l1->v_uval << l2->v_uval;break;
				/* Unsigned long modulo not supported */
				case TN_MOD :		l1->v_sval = l1->v_sval % l2->v_sval;break;
				case TN_MULTIPLY :	l1->v_uval = l1->v_uval * l2->v_uval;break;
				case TN_NOTEQUAL :	l1->v_sval = l1->v_uval != l2->v_uval;break;
				/* Unsigned long right shift not supported */
				case TN_RSHIFT :	l1->v_sval = l1->v_sval >> l2->v_sval;break;
				case TN_SUBTRACT :	l1->v_uval = l1->v_uval - l2->v_uval;break;
				default:
					panic("invalid op for eval const (unsigned)");
			}
			break;
	}
	kill_tree(l2);
	return l1;
}
