#include "def.h"
#include "nodes.h"
#include "symtab.h"
#include "types.h"
#include "error.h"
#include "nodetypes.h"

static char *sccsid = "@(#) typecheck2.c 1.10@(#)";

	/*
	 * typechecking of assignment and operation assignment nodes
	 */

	/* typecheck an assignment node */

struct treenode *check_assign(type,l1,l2,l3)
int type;
struct treenode *l1,*l2,*l3;
{
	struct treenode *node, *tmp;
	struct type *t1 = skiprefnodes(l1->vtype);
	struct type *t2 = skiprefnodes(l2->vtype);

	if(!islval(l1)) {
		errl(LVALUEREQD,l1);
		kill_tree(l1);
		return l2;
	}
	if( t1 == 0 ) panic("node without type in check_assign()");
	if( t2 == 0 ) panic("node without type in check_assign()");

	/* check the allowable cross product of types */

	switch(CROSS(t1->type,t2->type,8)) {
		case CROSS(TT_STRUCTHEADER,TT_STRUCTHEADER,8) :
		case CROSS(TT_UNIONHEADER,TT_UNIONHEADER,8) :
			if(!size(t1)) {
				err(INCOMPLETETYPE);
				kill_tree(l1);
				return l2;
			} else if(!typematch(t1,t2)) {
				err(INCOMPTYPES);
				kill_tree(l1);
				return l2;
			} else {
				node = make_plain(type,l1,l2,l3);
				node->vtype = l1->vtype;
				l1->vtype->refcount++;
				return node;
			}
			break;
		case CROSS(TT_POINTERTO,TT_ARRAYOF,8) :
		case CROSS(TT_POINTERTO,TT_FUNCTIONRETURNING,8) :
			t1 = skiprefnodes( t1->tn.next );
			t2 = skiprefnodes( t2->tn.next );
			if(!typematch(t1,t2)) {
				err(INCOMPTYPES);
				kill_tree(l1);
				return l2;
			}
			l2 = make_addressof(l2);
			break;
		case CROSS(TT_POINTERTO,TT_POINTERTO,8) :
			t1 = skiprefnodes( t1->tn.next );
			t2 = skiprefnodes( t2->tn.next );
			if(!typematch(t1,t2))
				err(PTRPTRCONV);
			break;
		case CROSS(TT_POINTERTO,TT_BASICTYPE,8) :
			if(l2->ntype == TN_VALUE && l2->format == NF_LEAF_CONST
				&& l2->v_sval == 0)
				break;
			else if(t2->tu.flags & TF_BIT32) {
				err(PTRINTCONV);
				break;
			}
			err(INCOMPTYPES);
			kill_tree(l1);
			return l2;
			break;
		case CROSS(TT_BASICTYPE,TT_POINTERTO,8) :
			err(PTRINTCONV); break;
		case CROSS(TT_BASICTYPE,TT_ENUMHEADER,8) :
		case CROSS(TT_BASICTYPE,TT_BASICTYPE,8) :
			if(!(t2->tu.flags & TF_VOID)) {
				if(t2->tu.flags & (TF_BIT8|TF_BIT16)) extend(&l2);
				break;
			}
			break;
		case CROSS(TT_ENUMHEADER,TT_ENUMHEADER,8) :
			if(!typematch(t1,t2)) err(ENUMMISMATCH);
			break;
		case CROSS(TT_ENUMHEADER,TT_BASICTYPE,8) :
			err(ENUMMISMATCH);
			extend(&l2);
			 break;
		default:
			err(INVALIDTYPEFOROP,"=");
			kill_tree(l1);
			return l2;
	}
	node = make_plain(type,l1,l2,l3);
	node->vtype = l1->vtype;
	l1->vtype->refcount++;
	node->lineno = linecnt;
	return node;
}

	/* typechecking of operation assignment nodes */

struct treenode *check_opassign(type,l1,l2,l3)
int type;
struct treenode *l1,*l2,*l3;
{
	struct treenode *node, *tmp;
	struct type *t1 = skiprefnodes(l1->vtype);
	struct type *t2 = skiprefnodes(l2->vtype);

	if(!islval(l1)) {
		errl(LVALUEREQD,l1);
		kill_tree(l1);
		return l2;
	}

	switch(CROSS(t1->type,t2->type,8)) {
		case CROSS(TT_POINTERTO,TT_BASICTYPE,8) :
			switch( type ) {
				case TN_ADDASSIGN:
				case TN_SUBASSIGN:
					if( !isvoid(t2) ) {
						extend(&l2);
						break;
					}
				default:
					err(INVALIDTYPEFOROP,
						type == TN_ANDASSIGN	? "&=" :
						type == TN_DIVASSIGN	? "/=" :
						type == TN_EXORASSIGN	? "^=" :
						type == TN_LSHIFTASSIGN	? "<<=" :
						type == TN_MODASSIGN	? "%=" :
						type == TN_MULTASSIGN	? "*=" :
						type == TN_ORASSIGN		? "|=" :
						type == TN_RSHIFTASSIGN	? ">>=" : "should not happen" );
					kill_tree(l1);
					return l2;
			}
			break;
		case CROSS(TT_BASICTYPE,TT_POINTERTO,8) :
			err(PTRINTCONV); break;
		case CROSS(TT_BASICTYPE,TT_ENUMHEADER,8) :
		case CROSS(TT_ENUMHEADER,TT_BASICTYPE,8) :
		case CROSS(TT_BASICTYPE,TT_BASICTYPE,8) :
			if(!(t2->tu.flags & TF_VOID)) {
				if(t2->tu.flags & (TF_BIT8|TF_BIT16))
					extend(&l2);
				break;
			}
			break;
		case CROSS(TT_ENUMHEADER,TT_ENUMHEADER,8) :
			if(!typematch(t1,t2)) err(ENUMMISMATCH);
			break;
		default:
			err(INVALIDTYPEFOROP,
				type == TN_ADDASSIGN	? "+=" :
				type == TN_ANDASSIGN	? "&=" :
				type == TN_DIVASSIGN	? "/=" :
				type == TN_EXORASSIGN	? "^=" :
				type == TN_LSHIFTASSIGN	? "<<=" :
				type == TN_MODASSIGN	? "%=" :
				type == TN_MULTASSIGN	? "*=" :
				type == TN_ORASSIGN		? "|=" :
				type == TN_RSHIFTASSIGN	? ">>=" :
				type == TN_SUBASSIGN	? "-=" : "should not happen" );
			kill_tree(l1);
			return l2;
	}
	node = make_plain(type,l1,l2,l3);
	node->vtype = l1->vtype;
	l1->vtype->refcount++;
	node->lineno = linecnt;
	return node;
}
