#include "def.h"
#include "types.h"
static char *sccsid = "@(#) typematch.c 1.23@(#)";

/*
 *  Check for matching types by wandering down the two type chains, skipping
 *  any type ref nodes, and comparing relevant ones.  This routine makes no
 *  allowances for types which may be promoted into the other type; the types
 *  must match exactly.  Array sizes are checked if given in both types.
 */
int typematch(t1,t2)
register struct type *t1, *t2;
{
	for(;;) {
		/*
		 *  First, get rid of junk (reference nodes) at the head of each
		 *  chain.
		 */
		t1 = skiprefnodes(t1);
		t2 = skiprefnodes(t2);
		/*
		 *  If the two pointers now point at the same type node, we definitely
		 *  have equivalence.
		 */
		if(t1 == t2) return 1;
		/*
		 *  If they point at nodes of different types, then we don't have
		 *  equivalence.
		 */
		if(t1->type != t2->type) return 0;
		/*
		 *  Handle special cases.
		 */
		switch(t1->type) {
			case TT_ARRAYOF:
				/*
				 *  Array:  If both sizes are given, they must be equal.
				 */
				if(t2->tu.array.numentries && t1->tu.array.numentries &&
				   t2->tu.array.numentries != t1->tu.array.numentries) return 0;
				break;
			case TT_BASICTYPE:
				/*
				 *  Basic types:  check that they are the same, and return
				 *  appropriate value.
				 */
				return (t1->tu.flags == t2->tu.flags);
				/*
				 *  All the following mean a mismatch.  Structures, unions,
				 *  and enumerations only match if they are defined by the
				 *  same type node, which is checked above.
				 */
			case TT_STRUCTHEADER:
			case TT_UNIONHEADER:
			case TT_ENUMHEADER:
				return 0;
		}
		t1 = t1->tn.next;
		t2 = t2->tn.next;
	}
}
