/* @(#) types.h 1.25@(#)
 *
 *  This header file defines the format for type expressions.  The following
 *  structure is used for each node in a type chain.
 */
struct type {
	short type;
	short refcount;
	union {
		struct type *next;
		struct member *members;
		struct enumerator *enumlist;
		char *name;
	} tn;
	union {
		struct {
			long numentries;	/* size of array */
			long sizeeach;		/* size of entry */
		} array;
		short flags;			/* basic type flags */
		struct {
			long size;
			short alignreq;
		} su;
	} tu;
};

/*
 *  The next two structures are used for structure field and enumerator
 *  lists, respectively.
 */
struct member {
	struct member *next;
	char *name;
	struct type *type;
	long offset;
};

struct enumerator {
	struct enumerator *next;
	char *name;
	long value;
};

/*
 *  Values for the "type" field of each node.
 */
#define TT_POINTERTO			1
#define TT_FUNCTIONRETURNING	2
#define TT_ARRAYOF				3
#define TT_TYPEREF				4
#define TT_STRUCTHEADER			5
#define TT_ENUMHEADER			6
#define TT_BASICTYPE			7
#define TT_UNIONHEADER          8

/*
 *  Values for the "flags" field in a BASICTYPE node.
 */
#define TF_BIT8					1
#define TF_BIT16				2
#define TF_BIT32				4
#define TF_VOID					8
#define TF_SIZEMASK				15

#define TF_UNSIGNED				16

/*
 *  Type qualifier flags.
 */
#define QF_CONST				32
#define QF_VOLATILE				64

/*
 *  Functions
 */
extern struct type *skiprefnodes();
extern struct type *make_sign();

/* macro to determine if type is void */

#define isvoid(tnode) (tnode->type == TT_BASICTYPE && \
	(tnode->tu.flags & TF_VOID))

/* macro to determine if type is pointer to void */

#define isvoidp(tnode) (tnode->type == TT_POINTERTO && \
	tnode->tn.next->type == TT_BASICTYPE && \
	(tnode->tn.next->tu.flags & TF_VOID))
